// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_basic_tackle.h"
#include "body_forestall_block.h"

#include "bhv_attacker_offensive_move.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_AttackerOffensiveMove::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: AttackerOffensiveMove. target=(%.2f %.2f)"
                        ,__FILE__, __LINE__,
                        M_home_pos.x, M_home_pos.y );

    //------------------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return true;;
    }

    //------------------------------------------------------
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: execute. intercept cycle. self=%d, mate=%d, opp=%d"
                        ,__FILE__, __LINE__,
                        self_min, mate_min, opp_min );
    //------------------------------------------------------
#if 1
    if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.6
         && opp_min < 3
         && opp_min < mate_min - 2
         && opp_min < self_min - 2
         && wm.ball().pos().dist( M_home_pos ) < 10.0 )
    {
        if ( Body_ForestallBlock( M_home_pos ).execute( agent ) )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            return true;
        }
    }
#endif
    //------------------------------------------------------
    if ( ( ! wm.existKickableTeammate() && self_min < 5 )
         || ( self_min < mate_min + 4 && mate_min > 3 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: execute. get ball"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "AttGetBall" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }

    //------------------------------------------------------
    // target point
    rcsc::Vector2D target_point = M_home_pos;

    if ( target_point.x > wm.getOffsideLineX() - 2.0 )
    {
        target_point.x = wm.getOffsideLineX() - 2.0;
    }

    const double dash_power = getDashPower( agent, target_point );

    ///////////////////////////////////////////////////

    agent->debugClient().addMessage( "AttackGo%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: execute. go to (%.2f, %.2f) dash_power=%.2f"
                        ,__FILE__, __LINE__,
                        target_point.x, target_point.y, dash_power );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.7 ) dist_thr = 0.7;

    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power,
                                 100, // cycle
                                 false, // back
                                 true, // stamina save
                                 40.0 // angle threshold
                                 ).execute( agent ) )
    {
        rcsc::Vector2D face_point( 100.0, wm.self().pos().y );
        if ( wm.ball().angleFromSelf().abs() > 100.0 )
        {
            face_point.x = wm.self().pos().x;
            face_point.y = wm.ball().pos().y;
        }
        rcsc::Body_TurnToPoint( face_point ).execute( agent );
    }
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

    return true;
}


/*-------------------------------------------------------------------*/
/*!

*/
double
Bhv_AttackerOffensiveMove::getDashPower( const rcsc::PlayerAgent * agent,
                                         const rcsc::Vector2D & target_point )
{
    const rcsc::WorldModel & wm = agent->world();
    //------------------------------------------------------
    // decide dash power
    static bool s_recover_mode = false;
    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.4 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: getDashPower. change to recover mode."
                            ,__FILE__, __LINE__ );
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: getDashPower. came back from recover mode"
                            ,__FILE__, __LINE__ );
        s_recover_mode = false;
    }

    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    // dash power
    if ( s_recover_mode )
    {
        // Magic Number.
        // recommended one cycle's stamina recover value
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: getDashPower. recover mode"
                            ,__FILE__, __LINE__ );
        return std::max( 0.0, my_inc - 20.0 );
    }

    if ( ! wm.getOpponentsFromSelf().empty()
         && wm.getOpponentsFromSelf().front()->distFromSelf() < 1.0 )
    {
        // opponent is very close
        // full power
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: getDashPower. exist near opponent. full power"
                            ,__FILE__, __LINE__ );
        return rcsc::ServerParam::i().maxPower();
    }

    if  ( wm.ball().pos().x < wm.self().pos().x
          && wm.self().pos().x < wm.getOffsideLineX() )
    {
        // ball is back
        // not offside
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: getDashPower. ball is back and not offside."
                            ,__FILE__, __LINE__ );
        if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.8 )
        {
            return std::min( my_inc * 0.9,
                             rcsc::ServerParam::i().maxPower() );
        }
        else
        {
            return std::min( my_inc * 1.1,
                             rcsc::ServerParam::i().maxPower() );
        }
    }

    if ( wm.ball().pos().x > wm.self().pos().x + 3.0 )
    {
        // ball is front, full power
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: getDashPower. ball is front full powerr"
                            ,__FILE__, __LINE__ );
        return rcsc::ServerParam::i().maxPower();
    }


    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: getDashPower. normal mode."
                        ,__FILE__, __LINE__ );
    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.8 )
    {
        return std::min( my_inc * 0.9,
                         rcsc::ServerParam::i().maxPower() );
    }
    else
    {
        return std::min( my_inc * 1.5,
                         rcsc::ServerParam::i().maxPower() );
    }
}
