// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>

#include "bhv_basic_defensive_kick.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_BasicDefensiveKick::execute( rcsc::PlayerAgent * agent )
{
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "Bhv_BasicDefensiveKick. opp dist=%.2f",
                        closest_opp_dist );

    const rcsc::WorldModel & wm = agent->world();
#if 1
    rcsc::Vector2D pass_point;

    if ( rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL) )
    {
        if ( pass_point.x > wm.self().pos().x + 20.0 )
        {
            if ( rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s: pass to front (%.2f %.2f )"
                                    ,__FILE__,
                                    pass_point.x, pass_point.y );
                agent->debugClient().addMessage( "DefPass1" );
                agent->setNeckAction( new rcsc::Neck_ScanField() );
                return true;
            }
        }
        if ( pass_point.x > wm.self().pos().x + 7,0 )
        {
            double opp_dist = 1000.0;
            const rcsc::PlayerObject * opp
                = wm.getOpponentNearestTo( pass_point,
                                           10,
                                           &opp_dist );
            if ( opp && opp_dist > 7.0 )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s: pass to front (%.2f %.2f)"
                                    ,__FILE__,
                                    pass_point.x, pass_point.y );
                agent->debugClient().addMessage( "DefPass1" );
                agent->setNeckAction( new rcsc::Neck_ScanField() );
                return true;
            }
        }
    }
#endif
    if ( closest_opp_dist > 12.0 )
    {
        rcsc::Vector2D target( 36.0, 0.0 );
        rcsc::Body_Dribble( target,
                            1.0,
                            rcsc::ServerParam::i().maxPower() * 0.7,
                            1,
                            false // never avoid
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s: dribble(1) to (%.2f %.2f)"
                            ,__FILE__,
                            target.x, target.y );
        agent->debugClient().addMessage( "DefDrib1" );
        return true;
    }

    // safety area
    if ( closest_opp_dist > 8.0
         && wm.self().pos().x > -30.0 )
    {
        rcsc::Vector2D target(36.0, 0.0);
        rcsc::Body_Dribble( target,
                            1.0,
                            rcsc::ServerParam::i().maxPower() * 0.5,
                            1,
                            false // never avoid
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s: dribble(2) to (%.2f, %.2f)"
                            ,__FILE__,
                            target.x, target.y );
        agent->debugClient().addMessage( "DefDrib2" );
        return true;
    }

    // can pass
    if ( rcsc::Body_Pass().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "DefPassDefault" );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    // dribble
    if ( closest_opp_dist > 12.0
         && wm.self().pos().absY() > 20.0 )
    {
        rcsc::AngleDeg ang_l, ang_r;
        if ( wm.self().pos().y < 0.0 )
        {
            ang_l = ( rcsc::Vector2D(0.0, 0.0) - wm.self().pos() ).th();
            ang_r = 80.0;
        }
        else
        {
            ang_l = -80.0;
            ang_r = ( rcsc::Vector2D(0.0, 0.0) - wm.self().pos() ).th();
        }

        if ( ang_l.isLeftOf( wm.self().body() )
             && wm.self().body().isLeftOf( ang_r ) )
        {
            rcsc::Vector2D drib_target
                = wm.self().pos()
                + rcsc::Vector2D::polar2vector( 2.0, wm.self().body() );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: dribble to body dir"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "DefDrib3" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                1,
                                false // never avoid
                                ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: opp too far. dribble to their side"
                            ,__FILE__,__LINE__ );
        agent->debugClient().addMessage( "DefDrib4" );
        rcsc::Vector2D target( 36.0, 0.0 );
        rcsc::Body_Dribble( target,
                            1.0,
                            rcsc::ServerParam::i().maxPower(),
                            1,
                            false // never avoid
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    if ( closest_opp_dist > 3.0 )
    {
        rcsc::Vector2D face_point( 100.0, 0.0 );
        rcsc::Body_HoldBall( true, face_point ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        agent->debugClient().addMessage( "DefHold" );
        return true;
    }

    if ( wm.self().pos().x < -35.0
         && wm.self().pos().absY() < 16.0)
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "Act_BasicDefensiveKick. %d: clear", __LINE__ );
        agent->debugClient().addMessage( "DefClear" );
        rcsc::Body_ClearBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "Act_BasicDefensiveKick. %d: advance", __LINE__ );
    agent->debugClient().addMessage( "DefAdvance" );
    rcsc::Body_AdvanceBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );

    return true;
}
