// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/geom/line_2d.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_stop_dash.h>
#include <rcsc/action/body_intercept.h>

#include "bhv_goalie_basic_move.h"

#include "bhv_goalie_chase_ball.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_GoalieChaseBall::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_GoalieChaseBall"
                        ,__FILE__, __LINE__ );
    //----------------------------------------------------------
    // It is necessary to deal with the situation
    //  which opponent keeps the ball or dribbles.

    /*
      const double opp_to_ball_dist
      = ( agent->world().getOpponentsFromBall().empty()
      ? 1000.0
      : agent->world().getOpponentsFromBall().front()->distFromBall() );
      if ( opp_to_ball_dist < rcsc::ServerParam::i().defaultKickableArea() + 0.2 )
      {
      rcsc::Vector2D target_point
      = agent->world().getOpponentsFromBall().front()->pos();
      target_point += agent->world().getOpponentsFromBall().front()->vel();
      rcsc::dlog.addText( rcsc::Logger::TEAM,
      "%s:%d: go to ball keeper point direct"
      ,__FILE__, __LINE__ );
      doGoToCatchPoint(act, target_point);
      return true;
      }
    */

    const rcsc::WorldModel & wm = agent->world();

    ////////////////////////////////////////////////////////////////////////
    // get active interception catch point

    rcsc::Vector2D my_int_pos = wm.getInterceptTable()->getSelfInterceptPoint();
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s: execute. intercept point=(%.2f %.2f)"
                        ,__FILE__,
                        my_int_pos.x, my_int_pos.y );

    double pen_thr = wm.ball().distFromSelf() * 0.1 + 1.0;
    if ( pen_thr < 1.0 ) pen_thr = 1.0;

    //----------------------------------------------------------
    const rcsc::Line2D ball_line( wm.ball().pos(), wm.ball().vel().th() );
    const rcsc::Line2D defend_line( rcsc::Vector2D( wm.self().pos().x, -10.0 ),
                                    rcsc::Vector2D( wm.self().pos().x, 10.0 ) );

    if ( my_int_pos.x > - rcsc::ServerParam::i().pitchHalfLength() - 1.0
         && my_int_pos.x < rcsc::ServerParam::i().ourPenaltyAreaLine() - pen_thr
         && my_int_pos.absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth() - pen_thr )
    {
        bool save_recovery = false;
        if ( ball_line.dist( wm.self().pos() ) < rcsc::ServerParam::i().catchableArea() )
        {
            save_recovery = true;
        }
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s: execute normal intercept"
                            ,__FILE__ );
        agent->debugClient().addMessage( "Intercept(0)" );
        rcsc::Body_Intercept( save_recovery ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }

    int self_goalie_min = wm.getInterceptTable()->getSelfReachCycle();
    int opp_min_cyc = wm.getInterceptTable()->getOpponentReachCycle();

    rcsc::Vector2D intersection;
    if ( ! ball_line.intersection( defend_line, &intersection )
         || ball_line.dist( wm.self().pos() ) < wm.self().catchableArea() * 0.8
         || intersection.absY() > rcsc::ServerParam::i().goalHalfWidth() + 3.0
         )
    {
        if ( ! wm.existKickableOpponent() )
        {
            if ( self_goalie_min <= opp_min_cyc + 2
                 && my_int_pos.x > -rcsc::ServerParam::i().pitchHalfLength() - 2.0
                 && my_int_pos.x < rcsc::ServerParam::i().ourPenaltyAreaLine() - pen_thr
                 && my_int_pos.absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth() - pen_thr
                 )
            {
                if ( rcsc::Body_Intercept( false ).execute( agent ) )
                {
                    rcsc::dlog.addText( rcsc::Logger::TEAM,
                                        "%s: execute normal interception"
                                        ,__FILE__ );
                    agent->debugClient().addMessage( "Intercept(1)" );
                    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
                    return true;
                }
            }

            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s: execute. ball vel has same slope to my body??"
                                " myvel-ang=%f body=%f. go to ball direct"
                                ,__FILE__,
                                wm.ball().vel().th().degree(),
                                wm.self().body().degree() );
            // ball vel angle is same to my body angle
            doGoToCatchPoint( agent, wm.ball().pos() );
            agent->debugClient().addMessage( "GoToCatch(1)" );
            return true;
        }
    }

    //----------------------------------------------------------
    // body is already face to intersection
    // only dash to intersection

    // check catch point
    if ( intersection.absX() > ( rcsc::ServerParam::i().pitchHalfLength()
                                 + wm.self().catchableArea() )
         || intersection.x > ( rcsc::ServerParam::i().ourPenaltyAreaLine()
                               - wm.self().catchableArea() )
         || intersection.absY() > ( rcsc::ServerParam::i().penaltyAreaHalfWidth()
                                    - wm.self().catchableArea() )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s: execute intersection(%.2f, %.2f) over range."
                            ,__FILE__,
                            intersection.x, intersection.y );
        if ( rcsc::Body_Intercept( false ).execute( agent ) )
        {
            agent->debugClient().addMessage( "Intercept(2)" );
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            return true;
        }
        else
        {
            return Bhv_GoalieBasicMove().execute( agent );
        }
    }

    //----------------------------------------------------------
    // check already there
    const rcsc::Vector2D my_inertia_final_pos
        = wm.self().pos()
        + wm.self().vel()
        / (1.0 - wm.self().playerType().playerDecay());
    double dist_thr = 0.2 + wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    // if already intersection point stop dash
    if ( my_inertia_final_pos.dist( intersection ) < dist_thr )
    {
        agent->debugClient().addMessage( "StopForChase" );
        rcsc::Body_StopDash( false ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }

    //----------------------------------------------------------
    // forward or backward

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: slide move. point=(%f, %f)"
                        ,__FILE__, __LINE__,
                        intersection.x, intersection.y );
    if ( wm.ball().pos().x > -35.0 )
    {
        if ( wm.ball().pos().y * intersection.y < 0.0 ) // opposite side
        {
            intersection.y = 0.0;
        }
        else
        {
            intersection.y *= 0.5;
        }
    }
    doGoToCatchPoint( agent, intersection );
    agent->debugClient().addMessage( "GoToCatch(2)" );
    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Bhv_GoalieChaseBall::doGoToCatchPoint( rcsc::PlayerAgent * agent,
                                       const rcsc::Vector2D & target_point )
{
    double dash_power;

    rcsc::Vector2D rel = target_point - agent->world().self().pos();
    rel.rotate( -agent->world().self().body() );
    rcsc::AngleDeg rel_angle = rel.th();
    const double angle_buf
        = std::fabs( rcsc::AngleDeg::atan2_deg( agent->world().self().catchableArea() * 0.9,
                                                rel.r() ) );

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: GoToCatchPoint. (%f, %f). angle_diff=%f. angle_buf=%f"
                        ,__FILE__, __LINE__,
                        target_point.x, target_point.y,
                        rel_angle.degree(), angle_buf );

    // forward dash
    if ( rel_angle.abs() < angle_buf )
    {
        dash_power = std::min( agent->world().self().stamina(),
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: forward dash"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "Forward" );
        agent->doDash( dash_power );
    }
    // back dash
    else if ( rel_angle.abs() > 180.0 - angle_buf )
    {
        dash_power = std::min( agent->world().self().stamina(),
                               rcsc::ServerParam::i().maxPower() * 2.0 );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: back dash"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "Back" );
        agent->doDash( -0.5 * dash_power );
    }
    // forward dash turn
    else if ( rel_angle.abs() < 90.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: turn %f for forward dash"
                            ,__FILE__, __LINE__,
                            rel_angle.degree() );
        agent->debugClient().addMessage( "F-Trun" );
        agent->doTurn( rel_angle );
    }
    else
    {
        rel_angle -= 180.0;
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: turn %f for back dash"
                            ,__FILE__, __LINE__,
                            rel_angle.degree() );
        agent->debugClient().addMessage( "B-Trun" );
        agent->doTurn( rel_angle );
    }

    agent->debugClient().addMessage( "GoToCatch" );
    agent->debugClient().setTarget( target_point );

    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_GoalieChaseBall::is_ball_chase_situation( const rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    if ( wm.getGameMode().type() != rcsc::GameMode::PlayOn )
    {
        return false;
    }

    int self_goalie_min = wm.getInterceptTable()->getSelfReachCycle();
    int opp_min_cyc = wm.getInterceptTable()->getOpponentReachCycle();

    ////////////////////////////////////////////////////////////////////////
    // ball is in very dangerous area
    const rcsc::Vector2D ball_next_pos = wm.ball().pos() + wm.ball().vel();
    if ( ball_next_pos.x < -rcsc::ServerParam::i().pitchHalfLength() + 8.0
         && ball_next_pos.absY() < rcsc::ServerParam::i().goalHalfWidth() + 3.0 )
    {
        // exist kickable teammate
        // avoid back pass
        if ( ! wm.getTeammatesFromBall().empty()
             && ( wm.getTeammatesFromBall().front()->distFromBall()
                  < rcsc::ServerParam::i().defaultKickableArea() )
             )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: danger area. exist kickable teammate?"
                                ,__FILE__, __LINE__ );
            return false;
        }
        else if ( wm.ball().distFromSelf() < 4.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: danger area. ball is very near."
                                ,__FILE__, __LINE__ );
            return true;
        }
        else
        {
            if ( self_goalie_min > opp_min_cyc + 3
                 && opp_min_cyc < 7 )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: danger area. opponent may get tha ball faster than me"
                                    ,__FILE__, __LINE__ );
                return false;
            }
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: danger area. chase ball"
                                ,__FILE__, __LINE__ );
            return true;
        }
    }

    ////////////////////////////////////////////////////////////////////////
    // check shoot moving
    if ( is_ball_shoot_moving( agent )
         && self_goalie_min < opp_min_cyc )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: shoot moving. chase ball"
                            ,__FILE__, __LINE__ );
        return true;
    }

    ////////////////////////////////////////////////////////////////////////
    // get active interception catch point

    const rcsc::Vector2D my_int_pos = wm.getInterceptTable()->getSelfInterceptPoint();

    double pen_thr = wm.ball().distFromSelf() * 0.1 + 1.0;
    if ( pen_thr < 1.0 ) pen_thr = 1.0;
    if ( my_int_pos.absY() > rcsc::ServerParam::i().penaltyAreaHalfWidth() - pen_thr
         || my_int_pos.x > rcsc::ServerParam::i().ourPenaltyAreaLine() - pen_thr )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: intercept point is out of penalty"
                            ,__FILE__, __LINE__ );
        return false;
    }

    ////////////////////////////////////////////////////////////////////////
    // Now, I can chase the ball
    // check the ball possessor

    if ( wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: exist kickable player"
                            ,__FILE__, __LINE__ );
        return false;
    }

    if ( opp_min_cyc <= self_goalie_min - 2 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: opponent reach the ball faster than me"
                            ,__FILE__, __LINE__ );
        return false;
    }

    const double my_dist_to_catch = wm.self().pos().dist( my_int_pos );

    double opp_min_dist = 10000.0;
    wm.getOpponentNearestTo( my_int_pos, 30, &opp_min_dist );

    if ( opp_min_dist < my_dist_to_catch - 2.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: opponent is nearer than me. my_dist=%.2f  opp_dist=%.2f"
                            ,__FILE__, __LINE__,
                            my_dist_to_catch, opp_min_dist );
        return false;
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: exist interception point. try chase."
                        ,__FILE__, __LINE__ );
    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_GoalieChaseBall::is_ball_shoot_moving( const rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel& wm = agent->world();

    if ( wm.ball().distFromSelf() > 30.0 )
    {
        return false;
    }

    if ( wm.ball().pos().x > -34.5 )
    {
        return false;
    }

    // check opponent kicker
    if ( wm.existKickableOpponent() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: check shoot moving. opponent kickable "
                            ,__FILE__, __LINE__ );
        return false;
    }
    else if ( wm.existKickableTeammate() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: check shoot moving. teammate kickable"
                            ,__FILE__, __LINE__ );
        return false;
    }

    if ( wm.ball().vel().absX() < 0.1 )
    {
        if ( wm.ball().pos().x < -46.0
             && wm.ball().pos().absY() < rcsc::ServerParam::i().goalHalfWidth() + 2.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: check shoot moving. bvel.x(%f) is ZERO. but near to goal"
                                ,__FILE__, __LINE__,
                                wm.ball().vel().x );
            return true;
        }
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: check shoot moving. bvel,x is small"
                            ,__FILE__, __LINE__ );
        return false;
    }


    const rcsc::Line2D ball_line( wm.ball().pos(), wm.ball().vel().th() );
    const double intersection_y = ball_line.getY( -rcsc::ServerParam::i().pitchHalfLength() );

    if ( std::fabs( ball_line.getB() ) > 0.1
         && std::fabs( intersection_y ) < rcsc::ServerParam::i().goalHalfWidth() + 2.0 )
    {
        if ( wm.ball().pos().x < -40.0
             && wm.ball().pos().absY() < 15.0 )
        {
            const rcsc::Vector2D end_point
                = wm.ball().pos()
                + wm.ball().vel() / ( 1.0 - rcsc::ServerParam::i().ballDecay());
            if ( wm.ball().vel().r() > 0.5 // 1.0
                 && end_point.x < -rcsc::ServerParam::i().pitchHalfLength() + 2.0 )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: shoot to Y(%f). ball_line a=%f, b=%f, c=%f"
                                    ,__FILE__, __LINE__,
                                    intersection_y,
                                    ball_line.getA(),
                                    ball_line.getB(),
                                    ball_line.getC() );
                return true;
            }
        }
    }


    return false;
}
