// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/soccer_math.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>

#include "body_forestall_block.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
 */
bool
Body_ForestallBlock::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s: Body_ForestallBlock",
                        __FILE__ );

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerObject * opp = wm.getInterceptTable()->getFastestOpponent();

    if ( ! opp )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: execute(). no opponent"
                            ,__FILE__ );
        return false;
    }

    rcsc::Vector2D block_point = opp->pos();

    if ( wm.self().pos().x > opp->pos().x )
    {
        block_point.x -= 5.0;
    }
    else
    {
        block_point.x -= 2.0;
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s: execute() target opp=%d(%,2f, %,2f) block_point=(%.1f %.1f)"
                        ,__FILE__,
                        opp->unum(), opp->pos().x, opp->pos().y,
                        block_point.x, block_point.y );

    agent->debugClient().addMessage( "forestall" );
    agent->debugClient().setTarget( block_point );

    double dash_power = wm.self().getSafetyDashPower( rcsc::ServerParam::i().maxPower() );

    rcsc::Body_GoToPoint( block_point, 1.0, dash_power,
                          5, // cycle
                          false, // no back dash
                          true, // save recovery
                          40.0 // dir_thr
                          ).execute( agent );
    return true;
}
