// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/math_util.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_cross_to_center.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_keep_shoot_chance.h"

#include "body_kick_to_corner.h"

#include "bhv_basic_tackle.h"

#include "bhv_basic_move.h"
#include "bhv_attacker_offensive_move.h"

#include "strategy.h"
#include "formation.h"
#include "formation_uva.h"

#include "role_center_forward.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::execute( rcsc::PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().getTeammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doKick( rcsc::PlayerAgent * agent )
{
    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
        rcsc::Body_AdvanceBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        break;
    case Strategy::BA_Danger:
        rcsc::Body_ClearBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_OffMidField:
        doMiddleAreaKick( agent );
        break;
    case Strategy::BA_Cross: // x>39, y>17
        doCrossAreaKick( agent );
        break;
    case Strategy::BA_ShootChance:
        doShootAreaKick( agent );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doMove( rcsc::PlayerAgent * agent )
{
    int ball_reach_step = 0;
    if ( ! agent->world().existKickableTeammate()
         && ! agent->world().existKickableOpponent() )
    {
        ball_reach_step
            = std::min( agent->world().getInterceptTable()->getTeammateReachCycle(),
                        agent->world().getInterceptTable()->getOpponentReachCycle() );
    }
    const rcsc::Vector2D base_pos = agent->world().ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, agent->world().getOffsideLineX() - 1.0 );
    }

    ///// delete here, if you do not use SBSP formation ////////

    if ( uva::Formations::instance().valid() )
    {
    if ( agent->world().getOurSide() == 1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() > 0 )
     {
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else 
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      }
    }
    if ( agent->world().getOurSide() == -1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() < 0 )
     {
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else 
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      }
   }
   }
   
    ///////////////////////////////////////////////////////

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y );

    switch ( Strategy::get_ball_area(base_pos) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_AttackerOffensiveMove( home_pos ).execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        if ( agent->world().ball().pos().x > 25.0 )
        {
            doGoToCrossPoint( agent, home_pos );
        }
        else
        {
            Bhv_AttackerOffensiveMove( home_pos ).execute( agent );
        }
        break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        doGoToCrossPoint( agent, home_pos );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doMiddleAreaKick( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doMiddleAreaKick"
                        ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "MiddleArea" );

    const rcsc::WorldModel & wm = agent->world();

    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );
    const rcsc::Vector2D closest_opp_pos
        = ( agent->world().getOpponentsFromSelf().empty()
            ? rcsc::Vector2D(-1000.0, 0.0)
            : agent->world().getOpponentsFromSelf().front()->pos() );

    rcsc::Vector2D pass_point;
    bool can_pass = rcsc::Body_Pass::get_best_pass( agent->world(),
                                                    & pass_point,
                                                    NULL,
                                                    NULL );
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d:    passable point (%.1f %.1f)"
                        ,__FILE__, __LINE__,
                        pass_point.x, pass_point.y );

    if ( can_pass
         && closest_opp_dist < 3.0 )
    {
        rcsc::Body_Pass().execute( agent );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "MidPass(1)" );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //------------------------------------------------
    // pass to side area
#if 1
    rcsc::Rect2D front_rect( wm.self().pos().x,
                             wm.self().pos().y - 5.0,
                             10.0, 10.0 );
    if ( can_pass
         && pass_point.x > wm.self().pos().x
         && pass_point.absY() > 20.0
         && wm.existOpponentIn( front_rect, 10, false ) ) // without goalie
    {
        agent->debugClient().addMessage( "MidPassSide(1)" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }
#endif
#if 1
    if ( can_pass
         && pass_point.x > wm.self().pos().x + 10.0
         && pass_point.absY() > 20.0 )
    {
        agent->debugClient().addMessage( "MidPassSide(2)" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }
#endif
    rcsc::Vector2D drib_target(50.0, 0.0);
    drib_target.y = ( agent->world().self().pos().y > 0.0
                      ? 5.0 : -5.0 );

    //------------------------------------------------
    // opp's is behind of me
    if ( closest_opp_pos.x < agent->world().self().pos().x + 1.0
         && closest_opp_dist > rcsc::ServerParam::i().defaultPlayerSize() * 2.0 + 0.2 )
    {
        rcsc::AngleDeg drib_angle = (drib_target - agent->world().self().pos()).th();
        const rcsc::Sector2D sector(agent->world().self().pos(),
                                    0.5, 15.0,
                                    drib_angle - 30.0, drib_angle + 30.0);
        // oppnent check with goalie
        if ( ! agent->world().existOpponentIn( sector, 10, true ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: fast dribble"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidDribFast" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                5
                                ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: slow dribble"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidDribSlow" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower() * 0.5,
                                5
                                ).execute( agent );
        }
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }

    //------------------------------------------------
    // check pass point
    if ( can_pass
         && pass_point.x > agent->world().getOffsideLineX() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: chance pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "MidPassOverOffside" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }

    //------------------------------------------------
    // opp is far from me
    if ( closest_opp_dist > 3.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: opp far. dribble"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "MidDribD1" );
        rcsc::Body_Dribble( drib_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower(),
                            1
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }

    //------------------------------------------------
    // opp is close
    // pass
    if ( can_pass )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "MidPassNorm" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }

    //------------------------------------------------
    // other
    // opponent may be close to me.
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: kick near side"
                        ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "MidToCorner" );
    Body_KickToCorner( (agent->world().self().pos().y < 0.0) ).execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doCrossAreaKick( rcsc::PlayerAgent * agent )
{
    agent->debugClient().addMessage( "CrossArea" );
    if ( Bhv_CrossToCenter::get_best_point( agent, NULL ) )
    {
        Bhv_CrossToCenter().execute( agent );
        return;
    }

    Bhv_BasicOffensiveKick().execute( agent );
}

namespace {

double
getNearestOpponentDistTo( const rcsc::WorldModel & wm,
                          const rcsc::Vector2D & point )
{
    double min_dist2 = 10000.0;
    const rcsc::PlayerPtrCont::const_iterator iend
        = wm.getOpponentsFromSelf().end();
    for ( rcsc::PlayerPtrCont::const_iterator it
              = wm.getOpponentsFromSelf().begin();
          it != iend;
          ++it )
    {
        double dtmp2 = (*it)->pos().dist2( point );
        if ( dtmp2 < min_dist2 )
        {
            min_dist2 = dtmp2;
        }
    }

    return std::sqrt( min_dist2 );
}

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doShootAreaKick( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doShootAreaKick"
                        ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "KeepChance" );
    const rcsc::WorldModel & wm = agent->world();

    //------------------------------------------------------------//
    // check pass parameters
    rcsc::Vector2D pass_point( rcsc::Vector2D::INVALID );
    bool can_pass = rcsc::Body_Pass::get_best_pass( wm,
                                                    & pass_point,
                                                    NULL,
                                                    NULL );
    double pass_point_opp_dist = 20.0;
    if ( can_pass )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doShootAreaKick. PassPoint(%.1f %.1f)"
                            ,__FILE__,
                            pass_point.x, pass_point.y );
        pass_point_opp_dist = getNearestOpponentDistTo( wm, pass_point );
    }

    //------------------------------------------------------------//
    // check very good pass
    if ( can_pass
         && pass_point_opp_dist > 5.0
         && pass_point.x > 40.0
         && pass_point.absY() < 10.0 )
    {
        agent->debugClient().addMessage( "ShotAreaPass(1)" );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doShootAreaKick. PassTo(%.1f %.1f)"
                            ,__FILE__,
                            pass_point.x, pass_point.y );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
        return;
    }

    //------------------------------------------------------------//
    // check cross paramters
    rcsc::Vector2D cross_point( rcsc::Vector2D::INVALID );
    bool can_cross = Bhv_CrossToCenter::get_best_point( agent, &cross_point );
    double cross_point_opp_dist = 20.0;
    if ( can_cross )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doShootAreaKick. CrossPoint(%.1f %.1f)"
                            ,__FILE__,
                            cross_point.x, cross_point.y );

        if ( cross_point.absY() > 10.0
             && cross_point.y * wm.self().pos().y > 0.0 // same side
             && cross_point.absY() > wm.self().pos().absY() )
        {
            can_cross = false;
            agent->debugClient().addMessage( "InvalidCross" );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doShootAreaKick. cross point is same side &outer. cancel"
                                ,__FILE__,
                                cross_point.x, cross_point.y );
        }
        if ( can_cross )
        {
            cross_point_opp_dist = getNearestOpponentDistTo( wm, cross_point );
        }
    }

    //------------------------------------------------------------//
    // check good cross chance
    if ( can_cross
         && cross_point_opp_dist > 3.0 )
    {
        agent->debugClient().addMessage( "ShotAreaCross(1)" );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doShootAreaKick. found good coross(%.1f %.1f)"
                            ,__FILE__,
                            cross_point.x, cross_point.y );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

    //------------------------------------------------------------//
    // get opponent parameters
    const rcsc::PlayerObject * nearest_opp
        = wm.getOpponentNearestToSelf( 10 );
    const double opp_dist = ( nearest_opp
                              ? nearest_opp->distFromSelf()
                              : 1000.0 );
    const rcsc::Vector2D opp_pos = ( nearest_opp
                                     ? nearest_opp->pos()
                                     : rcsc::Vector2D( -1000.0, 0.0 ) );
    //------------------------------------------------------------//
    // get goalie parameters
    const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
    rcsc::Vector2D goalie_pos( 100.0, 0.0 );
    double goalie_dist = 15.0;
    if ( opp_goalie )
    {
        goalie_pos = opp_goalie->pos();
        goalie_dist = opp_goalie->distFromSelf();
    }

    //------------------------------------------------------------//
    // check dribble chance
    if ( wm.self().pos().x < 46.5
         && ( opp_dist > 3.0 || opp_pos.x < wm.self().pos().x + 0.5 )
         && ( goalie_dist > 5.0
              || wm.self().pos().x < goalie_pos.x - 3.0 )
         )
    {
        bool goalie_front = false;
        if ( opp_goalie
             && opp_goalie->pos().x < 46.0
             && opp_goalie->pos().x < wm.self().pos().x + 7.0
             && std::fabs( opp_goalie->pos().y - wm.self().pos().y ) < 1.6 )
        {
            goalie_front = true;
        }

        if ( ! goalie_front )
        {
            const rcsc::Rect2D target_rect( wm.ball().pos().x + 0.5,
                                            wm.ball().pos().y - 1.4,
                                            2.5, 2.8 );
            // opponent check with goalie
            if ( ! wm.existOpponentIn( target_rect, 10, true ) )
            {
                rcsc::Vector2D drib_target( 47.0, wm.self().pos().y );
                if ( drib_target.absY() > 12.0 )
                {
                    drib_target.y = 12.0 * ( drib_target.y > 0.0 ? 1.0 : -1.0 );
                }

                if ( opp_goalie
                     && opp_goalie->pos().x > wm.self().pos().x
                     //&& opp_goalie->pos().y * wm.self().pos().y < 0.0
                     && std::fabs( opp_goalie->pos().y - wm.self().pos().y ) < 3.0
                     && opp_goalie->pos().absY() < wm.self().pos().absY() )
                {
                    drib_target.x = opp_goalie->pos().x - 3.0;
                }
                if ( wm.self().pos().x > drib_target.x - 0.8
                     && wm.self().pos().absY() > 9.0 )
                {
                    drib_target.y = 7.0 * ( drib_target.y < 0.0 ? -1.0 : 1.0 );
                }
                agent->debugClient().addMessage( "ShotAreaDrib" );
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s: doShootAreaKick. dribble to (%.1f %.1f)"
                                    ,__FILE__,
                                    drib_target.x, drib_target.y );
                rcsc::Body_Dribble( drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower(),
                                    1, // one dash
                                    false // no dodge
                                    ).execute( agent );
                if ( ! doCheckCrossPoint( agent ) )
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
                }
                return;
            }
        }
    }

#if 1
    //------------------------------------------------------------//
    // check dribble chance (2)
    if ( wm.self().pos().x > 44.0
         && goalie_dist > 4.0
         )
    {
        rcsc::Vector2D drib_target( 45.0, 0.0 );
        const rcsc::Rect2D target_rect( wm.self().pos().x - 2.0,
                                        ( wm.self().pos().y > 0.0
                                          ? wm.self().pos().y - 5.0
                                          : wm.self().pos().y ),
                                        4.0, 5.0 );
        // opponent check with goalie
        if ( ! wm.existOpponentIn( target_rect, 10, true ) )
        {
            //agent->debugClient().addLine( target_rect.topLeft(),
            //                              target_rect.topRight() );
            //agent->debugClient().addLine( target_rect.topRight(),
            //                              target_rect.bottomRight() );
            //agent->debugClient().addLine( target_rect.bottomRight(),
            //                              target_rect.bottomLeft() );
            //agent->debugClient().addLine( target_rect.bottomLeft(),
            //                              target_rect.topLeft() );

            if ( opp_goalie
                 && opp_goalie->pos().x > wm.self().pos().x
                 && std::fabs( opp_goalie->pos().y - wm.self().pos().y ) < 3.0
                 && opp_goalie->pos().absY() < wm.self().pos().absY() )
            {
                drib_target.x = opp_goalie->pos().x - 3.0;
            }
            agent->debugClient().addMessage( "ShotAreaDrib2" );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doShootAreaKick. dribble2 to (%.1f %.1f)"
                                ,__FILE__,
                                drib_target.x, drib_target.y );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower() * 0.7,
                                1, // one dash
                                false // no dodge
                                ).execute( agent );
            if ( ! doCheckCrossPoint( agent ) )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            }
            return;
        }
    }
#endif

    //------------------------------------------------------------//
    // check opponent
    // if opponent is close
    if ( opp_dist < 2.0
         || goalie_dist < 4.0 )
    {
        if ( can_cross )
        {
            agent->debugClient().addMessage( "ShotAreaCross(2)" );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doShootAreaKick. cross(2)  opp_dist=%.2f goalie_dist=%.2f"
                                ,__FILE__,
                                opp_dist, goalie_dist );
            Bhv_CrossToCenter().execute( agent );
            return;
        }
        if ( can_pass )
        {
            agent->debugClient().addMessage( "ShotAreaPass(2)" );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doShootAreaKick. pass  opp_dist=%.2f goalie_dist=%.2f"
                                ,__FILE__,
                                opp_dist, goalie_dist );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            return;
        }

        rcsc::Rect2D goal_area( rcsc::ServerParam::i().pitchHalfLength() - 10.0,
                                -10.0,
                                13.0, 20.0 );
        if ( wm.existTeammateIn( goal_area, 10, true ) )
        {
            rcsc::Vector2D kick_target( 45.0, 20.0 );

            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doShootAreaKick. enforce cross"
                                ,__FILE__ );

            if ( wm.self().pos().y > 0.0 ) kick_target.y *= -1.0;
            agent->debugClient().setTarget( kick_target );
            if ( goalie_dist < 3.0 )
            {
                agent->debugClient().addMessage( "ForceCross1K" );
                rcsc::Body_KickOneStep( kick_target,
                                        rcsc::ServerParam::i().ballSpeedMax()
                                        ).execute( agent );
            }
            else
            {
                agent->debugClient().addMessage( "ForceCross2K" );
                rcsc::Body_KickTwoStep( kick_target,
                                        rcsc::ServerParam::i().ballSpeedMax(),
                                        true ).execute( agent );
            }
            if ( ! doCheckCrossPoint( agent ) )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            }
            return;
        }
    }

    agent->debugClient().addMessage( "ShotAreaHold" );
    rcsc::Vector2D face_point( 52.5, 0.0 );
    rcsc::Body_HoldBall( true, face_point ).execute( agent );

    if ( ! doCheckCrossPoint( agent ) )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doGoToCrossPoint( rcsc::PlayerAgent * agent,
                                     const rcsc::Vector2D& home_pos )
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doGoToCrossPoint"
                        ,__FILE__, __LINE__ );
    //------------------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    //----------------------------------------------
    // intercept check
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();

    if ( self_min < mate_min
         || ( ! wm.existKickableTeammate()
              && self_min <= 6
              && wm.ball().pos().dist( home_pos ) < 10.0 )
         || wm.getInterceptTable()->isSelfFastestPlayer() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doGoToCross. get ball"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "ToCrossGetBall" );

        rcsc::Body_Intercept().execute( agent );
        if ( wm.self().pos().x > 30.0 )
        {
            if ( ! doCheckCrossPoint( agent ) )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            }
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return;
    }

    //----------------------------------------------
    // ball owner check
#if 0
    if ( wm.existKickableOpponent() )
    {
        if ( self_min <= 6
             && wm.ball().pos().dist( home_pos ) < 10.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: doGoToCross. get ball from opp"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "ToCrossGetBall" );
            ActB_Intercept().execute( agent );
            Act_NeckOnlyToBallOrScan().execute( agent );
            return;
        }

        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d:doGoToCross. opp kickable "
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "ToCrossOppOwn(1)" );
        Act_BasicMove( home_pos ).execute( agent );
        return;
    }
#endif
    if ( ! wm.getInterceptTable()->isOurTeamBallPossessor() )
    {
        const rcsc::PlayerObject * opp = wm.getOpponentNearestToBall( 3 );

        if ( opp &&  opp->distFromSelf() < 2.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: doGoToCross. opp has ball"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "ToCrossOppOwn(2)" );
            Bhv_BasicMove( home_pos ).execute( agent );
            return;
        }
    }

    //----------------------------------------------
    // set target
#if 0
    rcsc::Vector2D target_point( wm.ball().pos().x - 1.0, home_pos.y );
    if ( wm.ball().pos().absY() < 20.0 )
    {
        target_point.y = 5.0;
        if ( wm.ball().pos().y < 0.0 ) target_point.y *= -1.0;
    }
    if ( target_point.x > 44.0 ) target_point.x = 44.0;

    // consider near opponent
    if ( wm.self().pos().dist( target_point ) < 5.0 )
    {
        double opp_dist = 200.0;
        const PlayerObject * opp = wm.getOpponentNearestTo( target_point,
                                                            10,
                                                            &opp_dist );
        if ( opp && opp_dist < 2.5 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: ShootAreaMove. blocked. old target=(%.2f, %.2f)"
                                ,__FILE__, __LINE__,
                                target_point.x, target_point.y );
            agent->debugClient().addMessage( "Avoid" );
            target_point.x -= 3.0;
        }
    }
#else
    rcsc::Vector2D target_point = home_pos;
#endif

    //----------------------------------------------
    // set dash power
    // check X buffer & stamina
    static bool s_recover_mode = false;
    if ( wm.self().pos().x > 35.0
         && wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 500.0 )
    {
        s_recover_mode = true;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doGoToCross. recover on"
                            ,__FILE__, __LINE__ );
    }
    if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = false;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doGoToCross. recover off"
                            ,__FILE__, __LINE__ );
    }

    double dash_power = rcsc::ServerParam::i().maxPower();
    if ( s_recover_mode )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.6;
    }
    else if ( wm.ball().pos().x > wm.self().pos().x )
    {
        if ( wm.existKickableTeammate()
             && wm.self().pos().x > 30.0
             && wm.ball().pos().x > 35.0 )
        {
            dash_power *= 0.5;
        }
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.8;
    }

    //----------------------------------------------
    // positioning to make the cross course!!

    agent->debugClient().addMessage( "GoToCross%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doGoToCross. to (%.2f, %.2f)"
                        ,__FILE__, __LINE__,
                        target_point.x, target_point.y );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power,
                                 5, // cycle
                                 false, // no back dash
                                 true, // save recovery
                                 30.0 // dir thr
                                 ).execute( agent ) )
    {
        rcsc::Vector2D face_point( 100.0, wm.self().pos().y );
        rcsc::Body_TurnToPoint( face_point ).execute( agent );
    }

    if ( wm.self().pos().x > 30.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
RoleCenterForward::doCheckCrossPoint( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    if ( wm.self().pos().x < 35.0 )
    {
        return false;
    }

    const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
    if ( opp_goalie && opp_goalie->posCount() > 2 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doCheckCrossTarget  goalie should be checked"
                            ,__FILE__ );
        return false;
    }

    rcsc::Vector2D opposite_pole( 46.0, 7.0 );
    if ( wm.self().pos().y > 0.0 ) opposite_pole *= -1.0;

    rcsc::AngleDeg opposite_pole_angle = ( opposite_pole - wm.self().pos() ).th();


    if ( wm.getDirCount( opposite_pole_angle ) <= 1 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doCheckCrossTarget enough accuracy to angle %.1f"
                            ,__FILE__,
                            opposite_pole_angle.degree() );
        return false;
    }

    rcsc::AngleDeg angle_diff
        = agent->effector().queuedNextAngleFromBody( opposite_pole );
    if ( angle_diff.abs() > 100.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doCheckCrossPoint. want to face opposite pole,"
                            " but over view range. angle_diff=%.1f"
                            ,__FILE__,
                            angle_diff.degree() );
        return false;
    }


    agent->setNeckAction( new rcsc::Neck_TurnToPoint( opposite_pole ) );
    agent->debugClient().addMessage( "NeckToOpposite" );
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doCheckCrossPoint Neck to oppsite pole"
                        ,__FILE__, __LINE__ );
    return true;
}
