// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/math_util.h>
#include <rcsc/soccer_math.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_field.h>

#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_cross_to_center.h"
#include "bhv_keep_shoot_chance.h"

#include "bhv_basic_move.h"
#include "bhv_defender_basic_block_move.h"
#include "bhv_defender_get_ball.h"
#include "bhv_press.h"

#include "strategy.h"
#include "formation.h"
#include "formation_uva.h"

#include "role_defensive_half.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::execute( rcsc::PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().getTeammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );

    switch ( Strategy::get_ball_area(agent->world().ball().pos()) ) {
    case Strategy::BA_CrossBlock:
        if ( closest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: cross block area. danger"
                                ,__FILE__, __LINE__ );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::Body_AdvanceBall().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        if ( closest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: stopper or danger"
                                ,__FILE__, __LINE__ );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::Body_ClearBall().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicDefensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
        Bhv_CrossToCenter().execute( agent );
        break;
    case Strategy::BA_ShootChance:
        if ( Bhv_CrossToCenter::get_best_point( agent, NULL ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: shoot chance area"
                                ,__FILE__, __LINE__ );
            Bhv_CrossToCenter().execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: keep shoot chance"
                                ,__FILE__, __LINE__ );
            Bhv_KeepShootChance().execute( agent );
        }
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::Vector2D base_pos = agent->world().ball().pos();
    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, agent->world().getOffsideLineX() - 1.0 );
    }

    ///// delete here, if you do not use SBSP formation ////////

    if ( uva::Formations::instance().valid() )
    {
    if ( agent->world().getOurSide() == 1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() > 0 )
     {
      if (agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
     if (agent->world().ball().pos().x > 20.0 )
     {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
     }
     else if (agent->world().ball().pos().x < -15.0 )
     {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
     }
     else
     {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
     }
    }
    }
    if ( agent->world().getOurSide() == -1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() < 0 )
     {
      if (agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
     if (agent->world().ball().pos().x > 20.0 )
     {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
     }
     else if (agent->world().ball().pos().x < -15.0 )
     {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
     }
     else
     {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
     }
    }
   }
   }
    ///////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: HOME POSITION=(%.3f, %.3f)"
                        ,__FILE__, __LINE__,
                        home_pos.x, home_pos.y );

    switch ( Strategy::get_ball_area(base_pos) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
        doCrossBlockMove( agent, home_pos );
        break;
    case Strategy::BA_Danger:
        if ( agent->world().existKickableTeammate()
             && ! agent->world().existKickableOpponent() )
        {
            rcsc::Vector2D new_target = home_pos;
            new_target.x += 5.0;
            Bhv_BasicMove( new_target ).execute( agent );
        }
        else
        {
            Bhv_DefenderBasicBlockMove( home_pos ).execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
        doDribbleBlockMove( agent, home_pos );
        break;
    case Strategy::BA_DefMidField:
        doDefensiveMove( agent, home_pos );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( home_pos ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doCrossBlockMove( rcsc::PlayerAgent * agent,
                                     const rcsc::Vector2D& home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doCrossBlockMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    if ( wm.getInterceptTable()->isSelfFastestPlayer() )
    {
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( Bhv_DefenderGetBall( home_pos ).execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: get ball"
                            ,__FILE__, __LINE__ );
        return;
    }

    double dash_power;

    if ( wm.ball().distFromSelf() > 30.0 )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.9;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball is too far. dash_power=%f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    else if ( wm.ball().distFromSelf() > 20.0 )
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.5;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball far. dash_power=%f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.9;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: near. dash_power = %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }

    if ( wm.existKickableTeammate() )
    {
        dash_power = std::min( rcsc::ServerParam::i().maxPower() * 0.5,
                               dash_power );
    }

    dash_power = wm.self().getSafetyDashPower( dash_power );

    agent->debugClient().addMessage( "CrossBlock" );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    if ( ! rcsc::Body_GoToPoint( home_pos, dist_thr, dash_power
                                 ).execute( agent ) )
    {
        rcsc::Vector2D body_point( 100.0,
                                   wm.self().pos().y );
        if ( wm.ball().angleFromSelf().abs() > 80.0 )
        {
            body_point.x = -100.0;
        }
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }

    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doDefensiveMove( rcsc::PlayerAgent * agent,
                                    const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doDefensiveMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //////////////////////////////////////////////////////
    // tackle
    if ( wm.existKickableOpponent()
         && wm.self().tackleProbability() > 0.9 )
    {
        double tackle_power = rcsc::ServerParam::i().maxPower();
        if ( wm.self().body().abs() < 75.0 )
        {
            agent->doTackle( tackle_power );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
        else if ( wm.self().body().abs() > 105.0 )
        {
            agent->doTackle( -tackle_power );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
    }

    //----------------------------------------------
    // intercept
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();

    if ( self_min < mate_min )
    {
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( wm.getInterceptTable()->isSelfFastestPlayer() )
    {
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    //////////////////////////////////////////////////////
    if ( wm.getInterceptTable()->isOurTeamBallPossessor() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: opponent possesion"
                            ,__FILE__, __LINE__ );
        if ( std::fabs( wm.ball().pos().y - home_pos.y ) < 10.0
             && ( wm.ball().pos().x < wm.self().pos().x
                  || wm.ball().distFromSelf() < 10.0 )
             )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: ball near. press"
                                ,__FILE__, __LINE__ );
            Bhv_Press( home_pos ).execute( agent ); // include turn_neck
            return;
        }

        rcsc::Vector2D new_target( wm.ball().pos().x, home_pos.y );
        Bhv_BasicMove( new_target ).execute( agent );
        return;
    }


    //////////////////////////////////////////////////////
    // our team possession

    rcsc::Vector2D ball_future
        = rcsc::inertia_n_step_point( wm.ball().pos(),
                                      wm.ball().vel(),
                                      mate_min,
                                      rcsc::ServerParam::i().ballDecay() );

    const double future_x_diff
        = ball_future.x - wm.self().pos().x;

    double dash_power = rcsc::ServerParam::i().maxPower();

    if ( wm.existKickableTeammate()
         && wm.ball().distFromSelf() )
    {
        dash_power *= 0.5;
    }
    else if ( future_x_diff < 0.0 ) // ball is behind
    {
        dash_power *= 0.9;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.8 )
    {
        dash_power *= 0.8;
    }
    else
    {
        dash_power = wm.self().playerType().staminaIncMax();
        dash_power *= wm.self().recovery();
        dash_power *= 0.9;
    }

    // save recovery
    dash_power = wm.self().getSafetyDashPower( dash_power );

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: go to home. dash_power=%f"
                        ,__FILE__, __LINE__,
                        dash_power );
    agent->debugClient().addMessage(" DMFDefMove");

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    if ( ! rcsc::Body_GoToPoint( home_pos, dist_thr, dash_power
                                 ).execute( agent ) )
    {
        rcsc::Vector2D body_point( 100.0,
                                   wm.self().pos().y );
        if ( wm.ball().angleFromSelf().abs() > 80.0 )
        {
            body_point.x = wm.self().pos().x;
            body_point.y = ( wm.ball().pos().y > wm.self().pos().y
                             ? 100.0
                             : -100.0 );
        }
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doDribbleBlockMove( rcsc::PlayerAgent * agent,
                                       const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doDribbleBlockMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    ///////////////////////////////////////////////////
    // tackle
    if ( wm.existKickableOpponent()
         && wm.self().tackleProbability() > 0.75 )
    {
        double tackle_power = rcsc::ServerParam::i().maxPower();
        if ( wm.self().body().abs() < 80.0 )
        {
            agent->doTackle( tackle_power );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
        else if ( wm.self().body().abs() > 100.0 )
        {
            agent->doTackle( -tackle_power );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
    }

    ///////////////////////////////////////////////////
    if ( wm.getInterceptTable()->isSelfFastestPlayer() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: I am fastest. intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    ///////////////////////////////////////////////////
    double dash_power;

    if ( wm.self().pos().x + 5.0 < wm.ball().pos().x )
    {
        if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.7 )
        {
            dash_power
                = rcsc::ServerParam::i().maxPower() * 0.7
                * ( wm.self().stamina() / rcsc::ServerParam::i().staminaMax() );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dash_power=%f"
                                ,__FILE__, __LINE__,
                                dash_power );
        }
        else
        {
            dash_power
                = rcsc::ServerParam::i().maxPower() * 0.75
                - std::min( 30.0, wm.ball().distFromSelf() );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dash_power=%f"
                                ,__FILE__, __LINE__,
                                dash_power );
        }
    }
    else
    {
        dash_power
            = rcsc::ServerParam::i().maxPower()
            + 10.0
            - wm.ball().distFromSelf();
        dash_power = rcsc::min_max( 0.0, dash_power, rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: dash_power=%f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }



    if ( wm.existKickableTeammate() )
    {
        dash_power = std::min( rcsc::ServerParam::i().maxPower() * 0.5,
                               dash_power );
    }

    dash_power = wm.self().getSafetyDashPower( dash_power );

    agent->debugClient().addMessage( "drib block move" );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    if ( ! rcsc::Body_GoToPoint( home_pos, dist_thr, dash_power
                                 ).execute( agent ) )
    {
        // face to front or side
        rcsc::Vector2D body_point( 100.0,
                                   wm.self().pos().y );
        if ( wm.ball().angleFromSelf().abs() > 90.0 )
        {
            body_point.x = wm.self().pos().x;
            body_point.y = ( wm.ball().pos().y > wm.self().pos().y
                             ? 100.0
                             : -100.0 );
        }
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
}
