// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/soccer_math.h>
#include <rcsc/geom/line_2d.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_cross_to_center.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_keep_shoot_chance.h"

#include "bhv_basic_tackle.h"

#include "bhv_basic_move.h"
#include "bhv_defender_basic_block_move.h"
#include "bhv_danger_area_tackle.h"

#include "strategy.h"
#include "formation.h"
#include "formation_uva.h"

#include "role_side_back.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::execute( rcsc::PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().getTeammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doKick( rcsc::PlayerAgent * agent )
{
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );

    switch ( Strategy::get_ball_area(agent->world().ball().pos()) ) {
    case Strategy::BA_CrossBlock:
        if ( closest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: kick. danger"
                                ,__FILE__, __LINE__ );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::Body_AdvanceBall().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        if ( closest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: kick. very danger"
                                ,__FILE__, __LINE__ );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::Body_AdvanceBall().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicDefensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
        Bhv_CrossToCenter().execute( agent );
        //Act_NeckOnlyFieldScan().execute( agent );
        break;
    case Strategy::BA_ShootChance:
        Bhv_KeepShootChance().execute( agent );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::Vector2D base_pos = agent->world().ball().pos();
    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, agent->world().getOffsideLineX() - 1.0 );
    }

    ///// delete here, if you do not use SBSP formation ////////

    if ( uva::Formations::instance().valid() )
    {
    if ( agent->world().getOurSide() == 1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() > 0 )
     {  
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    }
    if ( agent->world().getOurSide() == -1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() < 0 )
     {  
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    }
    }
    ///////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: HOME POSITION=(%.3f, %.3f)"
                        ,__FILE__, __LINE__,
                        home_pos.x, home_pos.y );

    switch ( Strategy::get_ball_area(base_pos) ) {
    case Strategy::BA_CrossBlock:
        doCrossBlockAreaMove( agent, home_pos );
        break;
    case Strategy::BA_Stopper:
        Bhv_DefenderBasicBlockMove( home_pos ).execute( agent );
        break;
    case Strategy::BA_Danger:
        doDangerAreaMove( agent, home_pos );
        break;
    case Strategy::BA_DribbleBlock:
        // same side
        if ( agent->world().ball().pos().y * home_pos.y > 0.0
             && agent->world().self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.6 )
        {
            doBlockSideAttacker( agent, home_pos );
        }
        // opposite side
        else
        {
            doBasicMove( agent, home_pos );
        }
        break;
    case Strategy::BA_DefMidField:
        if ( agent->world().ball().pos().x < agent->world().self().pos().x + 15.0
             && agent->world().ball().pos().y * home_pos.y > 0.0 // same side
             && agent->world().ball().pos().absY() > 7.0
             && agent->world().self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.6
             )
        {
            doBlockSideAttacker( agent, home_pos );
        }
        else
        {
            doBasicMove( agent, home_pos );
        }
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        doBasicMove( agent, home_pos );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( home_pos ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doBasicMove( rcsc::PlayerAgent * agent,
                           const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doBasicMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.9, 60.0 ).execute( agent ) )
    {
        return;
    }


    //--------------S------------------------------------
    // intercept

    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: intercept cycle. self=%d, mate=%d, opp=%d"
                        ,__FILE__, __LINE__,
                        self_min, mate_min, opp_min );

    if ( self_min <= 3
         || ( self_min < 20
              && ( self_min < mate_min
                   || ( self_min < mate_min + 3 && mate_min > 3 ) )
              && self_min < opp_min + 2 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    /*--------------------------------------------------------*/
    // go to home position

    // ball is in very safety area (very front & our ball)
    if ( wm.ball().pos().x > wm.self().pos().x + 10.0
         && wm.getInterceptTable()->isOurTeamBallPossessor() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball is front and our team own ball"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage("SB:safety");
        Bhv_BasicMove( home_pos ).execute( agent );
        return;
    }

    agent->debugClient().addMessage( "SB:move" );

    // decide wait position
    // adjust to the defence line
    rcsc::Vector2D target_point = home_pos;
    if ( home_pos.x < wm.self().pos().x // must go back
         && wm.ball().pos().x > wm.self().pos().x + 3.0 )
    {
        // make static line
        // 0, -10, -20, ...
        double tmpx = std::floor( wm.ball().pos().x * 0.1 ) * 10.0 - 3.0;
        if ( tmpx > 0.0 ) tmpx = 0.0;
        if ( tmpx > home_pos.x + 5.0 ) tmpx = home_pos.x + 5.0;
        target_point.x = tmpx;
        if ( wm.getDefenseLineX() < target_point.x - 1.0 )
        {
            target_point.x = wm.getDefenseLineX() + 1.0;
        }

        target_point.y = home_pos.y;

        rcsc::Vector2D our_corner( -52.0, -33.0 );
        if ( wm.ball().pos().y > 0.0 ) our_corner.y *= -1.0;
        rcsc::Line2D through_line( wm.ball().pos(),
                                   our_corner );
        double through_block_y = through_line.getY( target_point.x );
        if ( through_block_y * wm.ball().pos().y > 0.0 // same side
             && std::fabs( through_block_y ) < wm.ball().pos().absY() // ball is inner side
             && std::fabs( through_block_y - home_pos.y ) < 5.0 )
        {
            target_point.y = through_block_y;
            agent->debugClient().addMessage( "ThroughBlock" );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doBasicMove. block through pass lien."
                                " home.y=%.1f  block_y=%.1f"
                                ,__FILE__,
                                home_pos.y, through_block_y  );
        }
        else
        {
            double home_y_diff = wm.ball().pos().y - home_pos.y;
            target_point.y += home_y_diff * 0.1;
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: correct target point to wait ball. y-correct=%.3f"
                                ,__FILE__, __LINE__,
                                home_y_diff );
        }

        agent->debugClient().addMessage( "Wait" );
    }

    // decide dash power
    double dash_power
        = Bhv_DefenderBasicBlockMove::get_dash_power( agent, target_point );
    // decide threshold
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    agent->debugClient().addMessage( "%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power ).execute( agent ) )
    {
        rcsc::Vector2D body_point( wm.self().pos().x, 0.0 );
        body_point.y = ( wm.ball().pos().y < wm.self().pos().y
                         ? -40.0
                         : 40.0 );
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }

    if ( wm.existKickableOpponent()
         || opp_min <= 3 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doCrossBlockAreaMove( rcsc::PlayerAgent * agent,
                                    const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doCrossBlockAreaMove"
                        ,__FILE__, __LINE__ );

    //-----------------------------------------------
    // tackle
    if ( Bhv_DangerAreaTackle( 0.8 ).execute( agent ) )
    {
        return;
    }

    const rcsc::WorldModel & wm = agent->world();

    //--------------------------------------------------
    // intercept
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    if ( self_min < mate_min && self_min < opp_min )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    //-----------------------------------------
    // set move point

    rcsc::Vector2D center( -41.5, 0.0 );

    int min_cyc = mate_min;
    if ( opp_min < min_cyc )
    {
        min_cyc = opp_min;
    }

    rcsc::Vector2D ball_future
        = rcsc::inertia_n_step_point( wm.ball().pos(),
                                      wm.ball().vel(),
                                      min_cyc,
                                      rcsc::ServerParam::i().ballDecay() );

    if ( ball_future.y * home_pos.y < 0.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: opposite side. basic move"
                            ,__FILE__, __LINE__ );
        doBasicMove( agent, home_pos );
        return;
    }

    if ( ball_future.x < -45.0 )
    {
        center.x = ball_future.x + 1.0;
        center.x = std::max( -48.0, center.x );
    }

    rcsc::Line2D block_line( ball_future, (center - ball_future).th() );

    rcsc::Vector2D block_point;
    block_point.y = ball_future.absY() - 10.0;
    block_point.y = std::max(block_point.y, 15.0);
    if ( ball_future.y < 0.0 ) block_point.y *= -1.0;

    block_point.x = block_line.getX(block_point.y);

    //------------------------------------------
    // set dash power
    double dash_power = rcsc::ServerParam::i().maxPower();
    if ( wm.self().pos().x < block_point.x + 5.0 )
    {
        dash_power -= wm.ball().distFromSelf() * 2.0;
        dash_power = std::max( 20.0, dash_power );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: cross block go to (%f, %f) dash_powe=%f"
                        ,__FILE__, __LINE__,
                        block_point.x, block_point.y, dash_power );
    agent->debugClient().addMessage( "cross block%.0f", dash_power );
    agent->debugClient().setTarget( block_point );

    double dist_thr = wm.ball().distFromSelf() * 0.07;
    if ( dist_thr < 0.8 ) dist_thr = 0.8;

    if ( ! rcsc::Body_GoToPoint( block_point,
                                 dist_thr,
                                 dash_power
                                 ).execute( agent )
         )
    {
        rcsc::Vector2D body_point( -100.0, wm.self().pos().y );
        if ( wm.ball().angleFromSelf().abs() < 80.0 )
        {
            body_point.x = 100.0;
        }
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }

    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doDangerAreaMove( rcsc::PlayerAgent * agent,
                                const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doDangerAreaMove"
                        ,__FILE__, __LINE__ );

    //-----------------------------------------------
    // tackle
    if ( Bhv_DangerAreaTackle().execute( agent ) )
    {
        return;
    }

    const rcsc::WorldModel & wm = agent->world();

    const bool ball_is_same_side = ( wm.ball().pos().y * home_pos.y > 0.0 );

    if ( ! wm.existKickableOpponent()
         && ! wm.existKickableTeammate()
         && wm.ball().vel().y < 0.0 )
    {
        rcsc::Line2D ball_line( wm.ball().pos(),
                                wm.ball().vel().th() );
        double goal_line_y = ball_line.getY( rcsc::ServerParam::i().pitchHalfLength() );
        if ( std::fabs( goal_line_y ) < rcsc::ServerParam::i().goalHalfWidth() + 1.5 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: very very danger. try intercept"
                                ,__FILE__, __LINE__ );

            if ( rcsc::Body_Intercept().execute( agent ) )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
                return;
            }
        }
    }

    //--------------------------------------------------
    // intercept
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    if ( ball_is_same_side
         && ! wm.existKickableTeammate()
         && ( opp_min > 5 || self_min < 7 ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball is same side. try get"
                            ,__FILE__, __LINE__ );
        if ( rcsc::Body_Intercept().execute( agent ) )
        {
            agent->debugClient().addMessage( "DangerGetBall(1)" );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
    }

    const double closest_opp_dist
        = ( wm.getOpponentsFromBall().empty()
            ? 1000.0
            : wm.getOpponentsFromBall().front()->distFromSelf() );
    if ( closest_opp_dist > 3.0
         && wm.getInterceptTable()->isSelfFastestPlayer() )
    {
        rcsc::Body_Intercept().execute( agent );
        agent->debugClient().addMessage( "DangerGetBall(2)" );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( ball_is_same_side
         && wm.ball().pos().x < -44.0
         && wm.ball().pos().absY() < 18.0
         && ! wm.existKickableTeammate() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball is same side. and very danger. try intercept"
                            ,__FILE__, __LINE__ );
        if ( rcsc::Body_Intercept().execute( agent ) )
        {
            agent->debugClient().addMessage( "DangerGetBall(3)" );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
    }

    //--------------------------------------------------
    if ( ball_is_same_side
         && wm.ball().pos().absY() > rcsc::ServerParam::i().goalHalfWidth() )
    {
        double dash_power
            = wm.self().getSafetyDashPower( rcsc::ServerParam::i().maxPower() );
        // I am behind of ball
        if ( wm.self().pos().x < wm.ball().pos().x )
        {
            dash_power *= std::min( 1.0, 7.0 / wm.ball().distFromSelf() );
            dash_power = std::max( 30.0, dash_power );
        }
        agent->debugClient().addMessage( "DangerGoHome%.0f", dash_power );
        double dist_thr = wm.ball().distFromSelf() * 0.07;
        if ( dist_thr < 0.5 ) dist_thr = 0.5;
        if ( ! rcsc::Body_GoToPoint( home_pos,
                                     dist_thr,
                                     dash_power
                                     ).execute( agent )
             )
        {
            rcsc::Body_TurnToBall().execute( agent );
        }
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    //--------------------------------------------------
    // block opposite side goal
    rcsc::Vector2D block_point( -49.0,
                                rcsc::ServerParam::i().goalHalfWidth() - 3.0 );
#if 1
    if ( wm.self().pos().x < -45.0 )
    {
        if ( ! ball_is_same_side
             && wm.existKickableOpponent()
             && wm.ball().pos().x < -40.0
             //&& wm.ball().pos().absY() < 12.5
             && 5.0 < wm.ball().pos().absY() )
        {
            block_point.y = 0.5;
        }
        if ( ball_is_same_side
             && wm.existKickableOpponent()
             && wm.ball().pos().x < -40.0
             && wm.ball().pos().absY() < 12.5
             && 6.0 < wm.ball().pos().absY() )
        {
            block_point.y = rcsc::ServerParam::i().goalHalfWidth() - 1.5;
        }
    }
#endif
    if ( home_pos.y < 0.0 ) block_point.y *= -1.0;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: block goal"
                        ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "DangerBlockPole" );
    agent->debugClient().setTarget( block_point );
    double dash_power
        = wm.self().getSafetyDashPower( rcsc::ServerParam::i().maxPower() );
#if 1
    if ( wm.self().pos().x < 47.0
         && wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        if ( wm.self().pos().dist( block_point ) > 0.8 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: block goal with back move"
                                ,__FILE__, __LINE__ );
            rcsc::Bhv_GoToPointLookBall( block_point,
                                         1.0,
                                         dash_power
                                         ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: already block goal(1)"
                                ,__FILE__, __LINE__ );
            rcsc::Vector2D face_point( wm.self().pos().x, 0.0 );
            rcsc::Body_TurnToPoint( face_point ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }
#endif
    if ( ! rcsc::Body_GoToPoint( block_point,
                                 1.0,
                                 dash_power,
                                 100, // cycle
                                 false, // no back mode
                                 true, // save recovery
                                 15.0 // dir thr
                                 ).execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: already block goal(2)"
                            ,__FILE__, __LINE__ );
        rcsc::Vector2D face_point( wm.self().pos().x, 0.0 );
        rcsc::Body_TurnToPoint( face_point ).execute( agent );
    }
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doBlockSideAttacker( rcsc::PlayerAgent * agent,
                                   const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doBlockSideAttacker"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.9, 60.0 ).execute( agent ) )
    {
        return;
    }

    //--------------------------------------------------
    // intercept

    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    if ( self_min < opp_min && self_min < mate_min )
    {
        agent->debugClient().addMessage( "BlockGetBall(1)" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return;
    }

    rcsc::Vector2D ball_get_pos
        = rcsc::inertia_n_step_point( wm.ball().pos(),
                                      wm.ball().vel(),
                                      self_min,
                                      rcsc::ServerParam::i().ballDecay() );

    if ( self_min < 5
         && wm.self().pos().x < ball_get_pos.x
         && wm.self().pos().x < wm.ball().pos().x - 0.2
         && ( std::fabs( wm.ball().pos().y - wm.self().pos().y ) < 1.5
              || wm.ball().pos().absY() < wm.self().pos().absY()
              )
         )
    {
        agent->debugClient().addMessage( "BlockGetBall(2)" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return;
    }

    if ( wm.self().pos().x < wm.ball().pos().x - 5.0
         || ( home_pos.x < wm.ball().pos().x
              && wm.self().pos().x < home_pos.x )
         )
    {
        agent->debugClient().addMessage( "BlockToBasic" );
        doBasicMove( agent, home_pos );
        return;
    }

    //--------------------------------------------------
    // block move

    rcsc::Vector2D block_point = home_pos;

    if ( wm.ball().pos().x < wm.self().pos().x )
    {
        agent->debugClient().addMessage( "SideBlockStatic" );

        block_point.x = -42.0;
        block_point.y = wm.self().pos().y;
        if ( wm.ball().pos().absY() < wm.self().pos().absY() - 0.5 )
        {
            agent->debugClient().addMessage( "Correct" );
            block_point.y = 10.0;
            if ( home_pos.y < 0.0 ) block_point.y *= -1.0;
        }
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: static move"
                            ,__FILE__, __LINE__ );
    }
    else
    {
        agent->debugClient().addMessage( "SideBlockDynamic" );

        block_point.y = wm.ball().pos().y;
        block_point.y += wm.ball().vel().y;
        if ( block_point.y > 0.0 ) block_point.y -= 1.2;
        else block_point.y += 1.2;

        double my_dist = wm.self().pos().dist( block_point );
        double ball_dist = wm.ball().pos().dist( block_point );
        if ( my_dist > ball_dist - 1.0 )
        {
            block_point.x = std::floor( wm.ball().pos().x * 0.1 ) * 10.0;
            if ( block_point.x > home_pos.x + 5.0 ) block_point.x = home_pos.x + 5.0;
        }
        else if ( my_dist < ball_dist - 4.0 )
        {
            block_point.x = wm.self().pos().x;
        }

        if ( block_point.absY() < home_pos.absY() - 5.0 )
        {
            agent->debugClient().addMessage( "Correct" );
            block_point.y = home_pos.y;
            if ( home_pos.y < 0.0 ) block_point.y += 5.0;
            else block_point.y -= 5.0;
        }
    }

    agent->debugClient().setTarget( block_point );

    //-----------------------------------------
    // set dash power
    double dash_power = dash_power = rcsc::ServerParam::i().maxPower();
    if ( wm.self().pos().x < wm.ball().pos().x - 8.0 )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.8;
        agent->debugClient().addMessage( "PowSave(1)%.0f",
                                         dash_power );
    }
    else if ( wm.self().pos().x < wm.ball().pos().x
              && wm.self().pos().absY() < wm.ball().pos().absY() - 1.0
              && wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.65 )
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.7;
        agent->debugClient().addMessage( "PowSave(2)%.0f",
                                         dash_power);
    }
    else
    {
        agent->debugClient().addMessage( "PowMax" );
    }

    // move

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.8 ) dist_thr = 0.8;

    if ( ! rcsc::Body_GoToPoint( block_point, dist_thr, dash_power ).execute( agent ) )
    {
        rcsc::Vector2D face_point( -48.0, 0.0 );
        rcsc::Body_TurnToPoint( face_point ).execute( agent );
    }

    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
}
