// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/math_util.h>
#include <rcsc/geom/circle_2d.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_cross_to_center.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"

#include "body_kick_to_corner.h"
#include "body_kick_to_front_space.h"

#include "bhv_basic_tackle.h"

#include "bhv_attacker_offensive_move.h"
#include "bhv_basic_move.h"
#include "bhv_block_ball_kicker.h"

#include "strategy.h"
#include "formation.h"
#include "formation_uva.h"

#include "role_side_forward.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::execute( rcsc::PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().getTeammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doKick( rcsc::PlayerAgent * agent )
{
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
        if ( ! rcsc::Body_Pass().execute( agent ) )
        {
            if ( closest_opp_dist < rcsc::ServerParam::i().tackleDist() )
            {
                rcsc::Body_AdvanceBall().execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            }
            else
            {
                Bhv_BasicDefensiveKick().execute( agent );
            }
        }
        break;
    case Strategy::BA_Danger:
        if ( closest_opp_dist < rcsc::ServerParam::i().tackleDist() )
        {
            rcsc::Body_ClearBall().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
        doBasicKick( agent );
        break;
    case Strategy::BA_OffMidField:
        doMiddleAreaKick( agent );
        break;
    case Strategy::BA_Cross:
        doCrossAreaKick( agent );
        break;
    case Strategy::BA_ShootChance:
        doShootAreaKick( agent );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doMove( rcsc::PlayerAgent * agent )
{
    int ball_reach_step = 0;
    if ( ! agent->world().existKickableTeammate()
         && ! agent->world().existKickableOpponent() )
    {
        ball_reach_step
            = std::min( agent->world().getInterceptTable()->getTeammateReachCycle(),
                        agent->world().getInterceptTable()->getOpponentReachCycle() );
    }
    const rcsc::Vector2D base_pos = agent->world().ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, agent->world().getOffsideLineX() - 1.0 );
    }

/*   ///// delete here, if you do not use SBSP formation ////////

    if ( uva::Formations::instance().valid() )
    {
    if ( agent->world().getOurSide() == 1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() > 0 )
     {
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    }
    if ( agent->world().getOurSide() == -1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() < 0 )
     {
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    }
    }
*/   ///////////////////////////////////////////////////////

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y );

    switch ( Strategy::get_ball_area(base_pos) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_AttackerOffensiveMove( home_pos ).execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_AttackerOffensiveMove( home_pos ).execute( agent );
        break;
    case Strategy::BA_Cross: // x>39, y>17
    case Strategy::BA_ShootChance: // x>39, y<17
        doShootAreaMove( agent, home_pos );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( home_pos ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doBasicKick( rcsc::PlayerAgent * agent )
{
    // !!!check stamina!!!
    static bool S_recover_mode = false;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doBasicKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 200.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.6 )
    {
        S_recover_mode = false;
    }

    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    rcsc::Vector2D nearest_opp_pos( -1000.0, 0.0 );
    if ( nearest_opp ) nearest_opp_pos = nearest_opp->pos();

    rcsc::Vector2D pass_point;

    if ( rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: exist pass point"
                            ,__FILE__, __LINE__ );
        double opp_to_pass_dist = 100.0;
        wm.getOpponentNearestTo( pass_point, 10, &opp_to_pass_dist );
        if ( pass_point.x > 35.0
#if 1
             && opp_to_pass_dist > 10.0
#endif
             )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: very challenging pass"
                                ,__FILE__, __LINE__ );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }

        if ( nearest_opp_pos.x > wm.self().pos().x - 0.5
             && nearest_opp_dist < 5.0
             && ( pass_point.x > wm.self().pos().x + 5.0
                  || pass_point.x > wm.self().pos().x - 3.0 )
             )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack. keep away & attack pass"
                                ,__FILE__, __LINE__ );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }

        if ( nearest_opp_dist < rcsc::ServerParam::i().tackleDist() + 0.7
             && ( pass_point.x > 30.0
                  || pass_point.x > wm.self().pos().x - 3.0 )
             )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack(2). keep away & attack pass"
                                ,__FILE__, __LINE__ );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }

        if ( S_recover_mode
             && ( pass_point.x > 30.0
                  || pass_point.x > wm.self().pos().x - 3.0 )
             )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack. pass for stamina recover"
                                ,__FILE__, __LINE__ );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }

        //if ( pass_point.x > wm.self().pos().x )
        if ( pass_point.x > wm.self().pos().x - 3.0 )
        {
            const rcsc::Sector2D sector( wm.self().pos(),
                                         0.5, 15.0,
                                         -30.0, 30.0 );
            // opponent check with goalie
            if ( wm.existOpponentIn( sector, 10, true ) )
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s:%d: dribble attack. forward pass to avoid opp"
                                    ,__FILE__, __LINE__ );
                rcsc::Body_Pass().execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                return;
            }
        }
    }

    // recover mode
    if ( S_recover_mode )
    {
        if ( nearest_opp_dist > 5.0 )
        {
            rcsc::Body_HoldBall( true, rcsc::Vector2D(36.0, wm.self().pos().y * 0.9 )//0.0)
                                 ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
        else if ( rcsc::Body_Pass().execute( agent ) )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack. exhaust kick to corner"
                                ,__FILE__, __LINE__ );
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
        return;
    }

    if ( nearest_opp_dist < rcsc::ServerParam::i().defaultPlayerSize() * 2.0 + 0.2
         || wm.existKickableOpponent() )
    {
        if ( rcsc::Body_Pass().execute( agent ) )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }

        if ( wm.self().pos().x < 25.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack. kick to near side"
                                ,__FILE__, __LINE__ );
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
        }
        else if ( wm.self().pos().x < 35.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack. kick to space for keep away"
                                ,__FILE__, __LINE__ );
            Body_KickToFrontSpace().execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: dribble attack. opponent very near. cross to center."
                                ,__FILE__, __LINE__ );
            Bhv_CrossToCenter().execute( agent );
            return;
        }
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    // wing dribble
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doBasicKick. dribble attack. normal dribble"
                        ,__FILE__, __LINE__ );
    doDribbleAttack( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doDribbleAttack( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doDribbleAttack"
                        ,__FILE__, __LINE__ );

    bool goalie_near = false;
    //--------------------------------------------------
    // goalie check
    const rcsc::PlayerObject * opp_goalie = agent->world().getOpponentGoalie();
    if ( opp_goalie
         && opp_goalie->pos().x > rcsc::ServerParam::i().theirPenaltyAreaLine()
         && opp_goalie->pos().absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth()
         && ( opp_goalie->distFromSelf()
              < rcsc::ServerParam::i().catchAreaLength()
              + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 3.5 )
         )
    {
        goalie_near = true;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: dribble. goalie close"
                            ,__FILE__, __LINE__ );
    }

    // decide target point
    const rcsc::Vector2D drib_target = getDribbleTarget( agent, goalie_near );
    rcsc::AngleDeg target_angle = ( drib_target - agent->world().self().pos() ).th();

    // decide dash count & looking point
    int dash_count = 5;
    if ( ! goalie_near )
    {
        dash_count = getDribbleAttackDashStep( agent, target_angle );
    }


    agent->debugClient().setTarget( drib_target );
    agent->debugClient().addMessage( "DribAtt%d", dash_count );

    rcsc::Body_Dribble( drib_target,
                        1.0,
                        rcsc::ServerParam::i().maxPower(),
                        dash_count
                        ).execute( agent );

    if ( agent->world().getDirCount( target_angle ) > 1 )
    {
        target_angle
            = ( drib_target - agent->effector().queuedNextMyPos() ).th();
        rcsc::AngleDeg next_body = agent->effector().queuedNextMyBody();
        if ( ( target_angle - next_body ).abs() < 90.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: SB Dribble LookDribTarget"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "LookDribTarget" );
            agent->setNeckAction( new rcsc::Neck_TurnToPoint( drib_target ) );
            return;
        }
    }

    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
}

/*-------------------------------------------------------------------*/
/*!

*/
rcsc::Vector2D
RoleSideForward::getDribbleTarget( const rcsc::PlayerAgent * agent,
                                   const bool goalie_near )
{
    const double base_x = 50.0; // old 45.0
    const double second_x = 40.0;
    const rcsc::WorldModel & wm = agent->world();

    rcsc::Vector2D drib_target( rcsc::Vector2D::INVALID );

    //--------------------------------------------------
    // goalie is close
    if ( goalie_near )
    {
        drib_target.assign( base_x, 25.0 );
        if ( wm.self().pos().y < 0.0 )
        {
            drib_target.y *= -1.0;
        }
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: goalie colse. normal target"
                            ,__FILE__, __LINE__ );

        return drib_target;
    }

    // goalie is safety

    // cross area
    if ( wm.self().pos().x > 45.0 )
    {
        const double cut_in_rect_l = rcsc::ServerParam::i().goalAreaLength();
        const double cut_in_rect_w
            = rcsc::ServerParam::i().penaltyAreaHalfWidth()
            - rcsc::ServerParam::i().goalAreaHalfWidth() - 2.0;
        const rcsc::Vector2D cut_in_size( cut_in_rect_l, cut_in_rect_w );
        rcsc::Vector2D cut_in_point( rcsc::ServerParam::i().pitchHalfLength()
                                     - cut_in_rect_l * 0.5 + 1.5,
                                     rcsc::ServerParam::i().penaltyAreaHalfWidth()
                                     - cut_in_rect_w * 0.5 + 1.0 );
        if ( wm.self().pos().y < 0.0)
        {
            cut_in_point.y *= -1.0;
        }

        const rcsc::Rect2D cut_in_rect( cut_in_point,
                                        cut_in_size.x, cut_in_size.y );
        // oponent check with goalie
        if ( ! wm.existOpponentIn( cut_in_rect, 10, true ) )
        {
            drib_target.assign( base_x, 12.0 );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: cut in (%.1f, %.1f)"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y );
        }
        else
        {
            drib_target.assign( second_x,
                                rcsc::min_max( 15.0,
                                               wm.self().pos().absY() - 1.0,
                                               25.0 ) );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: keep away (%.1f, %.1f)"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y );
        }
    }
    else if ( wm.self().pos().x > 36.0 )
    {
        if ( wm.self().pos().absY() < 5.0 )
        {
            drib_target.assign( base_x, 22.0 );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: keep away to (%f, %f)"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y );
        }
        else if ( wm.self().pos().absY() > 19.0 )
        {
            drib_target.assign( base_x, 12.0 );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: go their penalty area (%f, %f)"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y );
        }
        else
        {
            drib_target.assign( base_x, std::fabs( wm.self().pos().y ) );
            if ( drib_target.y > 30.0 ) drib_target.y = 27.0;
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: go my Y (%f, %f)"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y );
        }
    }
    // dribble attack area
    else
    {
        drib_target.assign( base_x, std::fabs(wm.self().pos().y) );
        if ( drib_target.y > 30.0 ) drib_target.y = 27.0;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: normal target (%f, %f)"
                            ,__FILE__, __LINE__,
                            drib_target.x, drib_target.y );
    }


    if ( wm.self().pos().y < 0.0 )
    {
        drib_target.y *= -1.0;
    }

    return drib_target;
}

/*-------------------------------------------------------------------*/
/*!

*/
int
RoleSideForward::getDribbleAttackDashStep( const rcsc::PlayerAgent * agent,
                                           const rcsc::AngleDeg & target_angle )
{
    const rcsc::Sector2D target_sector( agent->world().self().pos(),
                                        0.5, 10.0,
                                        target_angle - 30.0,
                                        target_angle + 30.0 );

    const int default_step = 5;
    double dash_dist = 4.0;;
    try {
        dash_dist = agent->world().self().playerType().getDashDistanceTable().at( default_step );
        dash_dist += agent->world().self().playerType().inertiaTravel
            ( agent->world().self().vel(), default_step ).r();
    }
    catch ( ... )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " Exception caught" << std::endl;
    }

    rcsc::Vector2D next_point( rcsc::Vector2D::POLE, dash_dist, target_angle );
    next_point += agent->world().self().pos();
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: getDribbleAttackDashStep. my_dash_dsit= %.2f  point(%.1f %.1f)"
                        ,__FILE__,
                        dash_dist, next_point.x, next_point.y  );

    const rcsc::PlayerPtrCont & opps = agent->world().getOpponentsFromSelf();
    const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
    bool exist_opp = false;
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
          it != opps_end;
          ++it )
    {
        if ( (*it)->posCount() >= 10 )
        {
            continue;
        }
        if ( ( (*it)->angleFromSelf() - target_angle ).abs() > 90.0 )
        {
            continue;
        }
        if ( (*it)->pos().dist( next_point ) <
             rcsc::PlayerTypeSet::i().getDummyType().realSpeedMax()
             * ( default_step + std::min( 4, (*it)->posCount() ) )
             - 2.0
             )
        {
            exist_opp = true;
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: getDribbleAttackDashStep. opp%d(%.1f %.1f) can reach my target point faster"
                                ,__FILE__,
                                (*it)->unum(),
                                (*it)->pos().x, (*it)->pos().y );
            break;
        }

        if ( (*it)->distFromSelf() > 10.0 )
        {
            break;
        }
        if ( target_sector.contains((*it)->pos()) )
        {
            exist_opp = true;
            break;
        }
    }

    if ( exist_opp )
    {
        return 1;
    }
    else if ( agent->world().self().pos().x > 36.0 )
    {
        //return 3;
        return 2;
    }

    return 5;

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doMiddleAreaKick( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doMiddleAreaKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );

    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                             ? nearest_opp->pos()
                                             : rcsc::Vector2D( -1000.0, 0.0 ) );

    rcsc::Vector2D pass_point;
    const bool is_passable
        = rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL);

    if ( is_passable
         && pass_point.x > wm.self().pos().x + 1.5 )
    {
        double opp_dist = 200.0;
        wm.getOpponentNearestTo( pass_point, 10, &opp_dist );
        if ( opp_dist > 5.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: first pass"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidPass(1)" );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
    }

    // passable & opp near
    if ( nearest_opp_dist < 3.0 )
    {
        if ( rcsc::Body_Pass().execute( agent ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: pass"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidPass(2)" );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
    }

    rcsc::Vector2D drib_target;

    if ( nearest_opp_dist >
         rcsc::ServerParam::i().tackleDist() + rcsc::ServerParam::i().defaultPlayerSpeedMax() )
    {
        // dribble to my body dir
        rcsc::AngleDeg ang_l, ang_r;
        if ( wm.self().pos().y > 0.0 )
        {
            ang_l = ( rcsc::Vector2D(52.5, -7.0) - wm.self().pos() ).th();
            ang_r = ( rcsc::Vector2D(52.5, 38.0) - wm.self().pos() ).th();
        }
        else
        {
            ang_l = ( rcsc::Vector2D(52.5, -38.0) - wm.self().pos() ).th();
            ang_r = ( rcsc::Vector2D(52.5, 7.0) - wm.self().pos() ).th();
        }

        if ( wm.self().body().isWithin( ang_l, ang_r ) )
        {
            drib_target
                = wm.self().pos() + rcsc::Vector2D::polar2vector(5.0, wm.self().body());
            if ( drib_target.absX() < rcsc::ServerParam::i().pitchHalfLength() - 1.0
                 && drib_target.absY() < rcsc::ServerParam::i().pitchHalfWidth() - 1.0
                 && ( wm.getDirRangeCount((drib_target - wm.self().pos()).th(), 2)
                      < 4 * (2 * 2 + 1) )
                 )
            {
                const rcsc::Sector2D sector(wm.self().pos(), 0.5, 10.0,
                                            wm.self().body() - 30.0,
                                            wm.self().body() + 30.0);
                // oponent check with goalie
                if ( ! wm.existOpponentIn( sector, 10, true ) )
                {
                    rcsc::dlog.addText( rcsc::Logger::ROLE,
                                        "%s:%d: dribble to may body dir"
                                        ,__FILE__, __LINE__ );
                    agent->debugClient().addMessage( "MidDribToBody(1)" );
                    rcsc::Body_Dribble( drib_target,
                                        1.0,
                                        rcsc::ServerParam::i().maxPower(),
                                        2
                                        ).execute( agent );
                    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                    return;
                }
            }
        }

    }

    const rcsc::Vector2D closest_opp_pos
        = ( wm.getOpponentsFromSelf().empty()
            ? rcsc::Vector2D(-100.0, 0.0)
            : wm.getOpponentsFromSelf().front()->pos() );

    drib_target.x = 50.0;
    drib_target.y = ( wm.self().pos().y > 0.0
                      ? 7.0
                      : -7.0 );
    const rcsc::AngleDeg target_angle = (drib_target - wm.self().pos()).th();
    const rcsc::Sector2D sector( wm.self().pos(),
                                 0.5, 15.0,
                                 target_angle - 30.0,
                                 target_angle + 30.0 );
    // opp's is behind of me
    if ( closest_opp_pos.x < wm.self().pos().x + 1.0
         && nearest_opp_dist > 2.0 )
    {
        // oponent check with goalie
        if ( ! wm.existOpponentIn( sector, 10, true ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: opp behind. dribble 3 dashes"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidDrib3D(1)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                3
                                ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: opp behind. dribble 1 dash"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidDrib1D(1)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                1
                                ).execute( agent );
        }
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }


    // opp is far from me
    if ( nearest_opp_dist > 8.0 )
    {
        // oponent check with goalie
        if ( ! wm.existOpponentIn( sector, 10, true ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: opp far. dribble 3 dashes"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidDrib3D(2)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                3
                                ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: opp far. dribble 1 dash"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidDrib1D(2)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower() * 0.8,
                                1
                                ).execute( agent );
        }
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    // opponent is close

    if ( nearest_opp_dist > 5.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: opp middle close. dribble"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "MidDribSlow" );
        rcsc::Body_Dribble( drib_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower() * 0.5,
                            1
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    // exist close opponent
    // pass
    if ( rcsc::Body_Pass().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: opp close. pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "MidPass" );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    // dribble to my body dir
    if ( nearest_opp_dist > rcsc::ServerParam::i().tackleDist()
         || nearest_opp_pos.x < wm.self().pos().x + 1.0 )
    {
        // dribble to my body dir
        if ( wm.self().body().abs() < 100.0 )
        {
            const rcsc::Vector2D body_dir_drib_target
                = wm.self().pos()
                + rcsc::Vector2D::polar2vector(5.0, wm.self().body());
            if ( body_dir_drib_target.x < rcsc::ServerParam::i().pitchHalfLength() - 1.0
                 && body_dir_drib_target.absY() < rcsc::ServerParam::i().pitchHalfWidth() - 1.0
                 && wm.getDirRangeCount(wm.self().body(), 2) < 3 * (2 * 2 + 1)
                 )
            {
                // check opponents
                // 10m, +-30 degree
                const rcsc::Sector2D body_dir_sector( wm.self().pos(),
                                                      0.5, 10.0,
                                                      wm.self().body() - 30.0,
                                                      wm.self().body() + 30.0);
                // oponent check with goalie
                if ( ! wm.existOpponentIn( sector, 10, true ) )
                {
                    rcsc::dlog.addText( rcsc::Logger::ROLE,
                                        "%s:%d: opp close. dribble to my body"
                                        ,__FILE__, __LINE__ );
                    agent->debugClient().addMessage( "MidDribToBody(2)" );
                    rcsc::Body_Dribble( body_dir_drib_target,
                                        1.0,
                                        rcsc::ServerParam::i().maxPower(),
                                        2
                                        ).execute( agent );
                    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                    return;
                }
            }
        }
    }

    if ( wm.self().pos().x > wm.getOffsideLineX() - 5.0 )
    {
        // kick to corner
        rcsc::Vector2D corner_pos( rcsc::ServerParam::i().pitchHalfLength() - 8.0,
                                   rcsc::ServerParam::i().pitchHalfWidth() - 8.0 );
        if ( wm.self().pos().y < 0.0 )
        {
            corner_pos.y *= -1.0;
        }

        // near side
        if ( wm.self().pos().x < 25.0
             || wm.self().pos().absY() > 18.0 )
        {
            const rcsc::Sector2D front_sector( wm.self().pos(),
                                               0.5, 10.0,
                                               -30.0, 30.0 );
            // oponent check with goalie
            if ( ! wm.existOpponentIn( front_sector, 10, true ) )
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s:%d: kick to corner"
                                    ,__FILE__, __LINE__ );
                agent->debugClient().addMessage( "MidToCorner(1)" );
                Body_KickToCorner( (wm.self().pos().y < 0.0) ).execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                return;
            }
        }

        const rcsc::AngleDeg corner_angle = (corner_pos - wm.self().pos()).th();
        const rcsc::Sector2D corner_sector( wm.self().pos(),
                                            0.5, 10.0,
                                            corner_angle - 15.0,
                                            corner_angle + 15.0 );
        // oponent check with goalie
        if ( ! wm.existOpponentIn( corner_sector, 10, true ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: kick to near side"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "MidToCorner(2)" );
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
    }

#if 1
    agent->debugClient().addMessage( "MidHold" );
    rcsc::Body_HoldBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
#else
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: advance"
                        ,__FILE__, __LINE__ );
    Act_AdvanceBall().execute( agent );
    Act_NeckOnlyToLowConfTeammate().execute( agent );
#endif
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doCrossAreaKick( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doCrossAreaKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //-------------------------------------------------------
    // !!!check stamina!!!
    static bool S_recover_mode = false;
    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 200.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.6 )
    {
        S_recover_mode = false;
    }

    //-------------------------------------------------------
    // check opponent fielder & goalie distance
    const double opp_dist
        = ( wm.getOpponentsFromSelf().empty()
            ? 1000.0
            : wm.getOpponentsFromSelf().front()->distFromSelf() );

    const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
    bool goalie_near = false;
    bool goalie_verynear = false;
    if ( opp_goalie )
    {
        if ( opp_goalie->pos().x > rcsc::ServerParam::i().theirPenaltyAreaLine()
             && opp_goalie->pos().absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth() )
        {
            if ( opp_goalie->distFromBall()
                 < ( rcsc::ServerParam::i().catchAreaLength()
                     + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 4.0 ) )
            {
                goalie_near = true;
            }
            if ( opp_goalie->distFromBall()
                 < ( rcsc::ServerParam::i().catchAreaLength()
                     + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 2.0 ) )
            {
                goalie_verynear = true;
            }
        }
    }

    //-------------------------------------------------------
    if ( goalie_verynear )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: goalie very near. keep away from goalie "
                            ,__FILE__, __LINE__ );
        // kick to the near side corner
        agent->debugClient().addMessage( "XAreaGKNear" );
        Body_KickToCorner( (wm.self().pos().y < 0.0)
                           ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------

    rcsc::Vector2D pass_point;
    const bool can_pass = rcsc::Body_Pass::get_best_pass( wm, &pass_point,
                                                          NULL, NULL );
#if 1
    if ( can_pass
         && pass_point.x > 42.0
         && pass_point.absY() < 12.0
         && ! wm.existOpponentIn( rcsc::Circle2D( pass_point, 4.0 ), 10, true ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: CrossAreaKick. very chance pass."
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPass_0" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }
#endif
    if ( can_pass
         && pass_point.x > 39.0
         && pass_point.absY() < 7.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: can cross pass."
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPass1" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    // very side attack
    if ( wm.self().pos().absY() > 27.0 )
    {
        if ( can_pass
             && opp_dist < 4.0
             && pass_point.x > 26.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: very side attack.  cross"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "XAreaPass2" );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
    }

    //-------------------------------------------------------
    // cross to center
    // exist teammate at cross area
    if ( opp_dist < 3.0
         && Bhv_CrossToCenter::get_best_point( agent, NULL ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: teammate in cross area. cross to center"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaCross1" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

    //-------------------------------------------------------
    if ( S_recover_mode
         && ! goalie_near
         && opp_dist > 4.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaHold" );
        rcsc::Body_HoldBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------
    // keep dribble
    if ( ! goalie_near
         && opp_dist > rcsc::ServerParam::i().tackleDist() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: goalie not nea and not tackle opp. wing dribble"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaDrib1" );
        doDribbleAttack( agent );
        return;
    }
    //-------------------------------------------------------
    // forcely pass
    if ( can_pass )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: enforce pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPass3" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------
    // not exist opp on my body dir
    rcsc::Sector2D body_sector( wm.self().pos(),
                                0.5, 8.0,
                                wm.self().body() - 25.0,
                                wm.self().body() + 25.0 );
    // oponent check with goalie
    if ( ! wm.existOpponentIn( body_sector, 10, true ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: exsit opp in my body dir. wing dribble"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaDrib2" );
        doDribbleAttack( agent );
        return;
    }

    //-------------------------------------------------------
    // forcely cross
    if ( S_recover_mode && opp_dist < 2.5 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: enforce cross"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaCross2" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

    doDribbleAttack( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doShootAreaKick( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doShootAreaKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    //--------------------------------------------------------//
    // get pass info
    rcsc::Vector2D pass_point;
    const bool can_pass
        = rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL );

    if ( can_pass
         && pass_point.x > 44.0
         && pass_point.absY() < rcsc::ServerParam::i().goalHalfWidth() + 3.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ShootAreaKick Pass (%f, %f)"
                            ,__FILE__, __LINE__,
                            pass_point.x, pass_point.y );
        agent->debugClient().addMessage( "ShootAreaPass" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //--------------------------------------------------------//
    // get cross info
    rcsc::Vector2D cross_point;
    const bool can_cross
        = Bhv_CrossToCenter::get_best_point( agent, &cross_point );
    if ( can_cross
         && cross_point.x > 43.0
         && cross_point.absY() < 10.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ShootAreaKick Cross (%f, %f)"
                            ,__FILE__, __LINE__,
                            cross_point.x, cross_point.y );
        agent->debugClient().addMessage( "ShootAreaCross" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

    //--------------------------------------------------------//
    // try dribble
    rcsc::Vector2D cut_in_point( 50.0, rcsc::ServerParam::i().goalHalfWidth() );
    if ( wm.self().pos().y < 0.0 )
    {
        cut_in_point.y *= -1.0;
    }

    // check opponent near to dribble target
    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );
    const rcsc::Vector2D closest_opp_pos = ( nearest_opp
                                             ? nearest_opp->pos()
                                             : rcsc::Vector2D( -1000.0, 0.0 ) );

    if ( wm.self().pos().x > 45.0
         && wm.self().pos().absY() < 15.0 )
    {
        if ( nearest_opp->distFromSelf() > 2.0 )
        {
            rcsc::Vector2D face_point( 50.0, 0.0 );
            rcsc::Body_HoldBall( true, face_point ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            return;
        }
    }

    //------------------------------------------------------------//
    // check dribble chance
    if ( wm.self().pos().x > 38.0
         && wm.self().pos().absY() < 14.0 )
    {
        const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
        bool goalie_front = false;
        if ( opp_goalie
             && opp_goalie->pos().x < 46.0
             && opp_goalie->pos().x < wm.self().pos().x + 7.0
             && std::fabs( opp_goalie->pos().y - wm.self().pos().y ) < 1.6 )
        {
            goalie_front = true;
        }
        if ( ! goalie_front )
        {
            const rcsc::Rect2D target_rect( wm.ball().pos().x + 0.5,
                                            wm.ball().pos().y - 1.4,
                                            2.5,
                                            2.8 );
            // opponent check with goalie
            if ( ! wm.existOpponentIn( target_rect, 10, true ) )
            {
                rcsc::Vector2D drib_target( 47.0, wm.self().pos().y );
                if ( opp_goalie
                     && opp_goalie->pos().x > wm.self().pos().x
                     && opp_goalie->pos().y * wm.self().pos().y < 0.0
                     && opp_goalie->pos().absY() < wm.self().pos().absY() )
                {
                    drib_target.x = opp_goalie->pos().x - 3.0;
                }
                if ( wm.self().pos().x > drib_target.x - 0.8
                     && wm.self().pos().absY() > 9.0 )
                {
                    drib_target.y = 10.0 * ( drib_target.y < 0.0 ? -1.0 : 1.0 );
                }
                agent->debugClient().addMessage( "ShotAreaDrib" );
                rcsc::Body_Dribble( drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower(),
                                    1, // one dash
                                    false // no dodge
                                    ).execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
                return;
            }
        }
    }

    if ( cut_in_point.dist2( closest_opp_pos ) > 2.0 * 2.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ShootAreaKick opp not near. cut in dribble"
                            ,__FILE__, __LINE__ );
        doDribbleAttack( agent );
        return;
    }

    if ( can_cross )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ShootAreaKick  cross"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "ShootAreaCross2" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

    // pass
    if ( can_pass
         && pass_point.x > 30.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ShootAreaKick pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "ShootAreaPass2" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: ShootAreaKick Hold"
                        ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "ShootAreaHold" );
    rcsc::Body_HoldBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
}

/*-------------------------------------------------------------------*/
/*!

*/
double
RoleSideForward::getShootAreaMoveDashPower( rcsc::PlayerAgent * agent,
                                            const rcsc::Vector2D & target_point )
{
    const rcsc::WorldModel & wm = agent->world();

    static bool s_recover = false;

    double dash_power = 100.0;

    // recover check
    // check X buffer & stamina
    if ( wm.self().stamina() < rcsc::ServerParam::i().effortDecThrValue() + 500.0 )
    {
        if ( wm.self().pos().x > 30.0 )
        {
            s_recover = true;
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: recover on"
                                ,__FILE__, __LINE__ );
        }
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().effortIncThrValue() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: recover off"
                            ,__FILE__, __LINE__ );
        s_recover = false;
    }

    if ( wm.ball().pos().x > wm.self().pos().x )
    {
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( wm.ball().pos().x > 41.0
              && wm.ball().pos().absY() < rcsc::ServerParam::i().goalHalfWidth() + 5.0 )
    {
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.6;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.75;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball is far"
                            ,__FILE__, __LINE__ );
    }

    return dash_power;
}

/*-------------------------------------------------------------------*/
/*!

*/
rcsc::Vector2D
RoleSideForward::getShootAreaMoveTarget( rcsc::PlayerAgent * agent,
                                         const rcsc::Vector2D & home_pos )
{
    const rcsc::WorldModel & wm = agent->world();
#if 0
    // at first, set static position
    rcsc::Vector2D target_point( wm.ball().pos().x + 2.0,
                                 rcsc::ServerParam::i().goalHalfWidth() - 1.5 );
    if ( target_point.x > wm.getOffsideLineX() - 1.0 )
    {
        target_point.x = wm.getOffsideLineX() - 1.0;
    }
    if ( target_point.x > 49.0 ) target_point.x = 49.0;
    if ( wm.ball().pos().x > 45.0 ) target_point.y = 5.0;
    if ( home_pos.y < 0.0 ) target_point.y *= -1.0;

    // consider ball position
    if ( home_pos.y > 0.0 )
    {
        // if ball is inside than home position
        //    and ball is outside than target
        // correct target point Y to default home position
        if ( wm.ball().pos().y < home_pos.y
             && wm.ball().pos().y > target_point.y )
        {
            target_point.y = home_pos.y;
            agent->debugClient().addMessage( "AdjustY" );
        }
        // if ball is too far.
        if ( wm.ball().pos().y < -25.0 )
        {
            target_point.y = home_pos.y;
            agent->debugClient().addMessage( "TooFar" );
        }
    }
    else
    {
        if ( wm.ball().pos().y > home_pos.y
             && wm.ball().pos().y < target_point.y )
        {
            target_point.y = home_pos.y;
            agent->debugClient().addMessage( "AdjustY" );
        }
        if ( wm.ball().pos().y > 25.0 )
        {
            target_point.y = home_pos.y;
            agent->debugClient().addMessage( "TooFar" );
        }
    }

    if ( wm.ball().pos().y * home_pos.y < 0.0 // opposite side
         && wm.ball().pos().x > 45.0 // very chance
         && 6.0 < wm.ball().pos().absY()
         && wm.ball().pos().absY() < 15.0 )
    {
        Rect2D goal_area( 52.5 - 8.0, 52.5,
                          -6.0, 6.0 );
        if ( ! wm.existTeammateIn( goal_area, 10, true ) )
        {
            agent->debugClient().addMessage( "XSupport" );

            target_point.x = wm.ball().pos().x + 2.0;
            if ( target_point.x > wm.getOffsideLineX() - 1.0 )
            {
                target_point.x = wm.getOffsideLineX() - 1.0;
            }
            if ( target_point.x > 49.0 ) target_point.x = 49.0;

            target_point.y = 1.0;
            if ( home_pos.y < 0.0 ) target_point.y *= -1.0;
        }
    }

    // if ball is front,
    // consider self body direction
    if ( target_point.x > wm.self().pos().x + 3.0 //wm.self().pos().x < 36.0
         && wm.ball().pos().x > wm.self().pos().x + 2.0 )
    {
        rcsc::Vector2D goal_pos( 52.5, 0.0 );
        rcsc::AngleDeg goal_dir = ( goal_pos - wm.self().pos() ).th();
        bool already_facing = false;
        if ( wm.self().pos().y > 0.0 )
        {
            rcsc::Vector2D r_limit( 52.5, 25.0 );
            rcsc::AngleDeg r_dir = ( r_limit - wm.self().pos() ).th();
            if ( goal_dir.degree() < wm.self().body().degree()
                 && wm.self().body().degree() < r_dir.degree() )
            {
                already_facing = true;;
            }
        }
        else
        {
            rcsc::Vector2D l_limit( 52.5, -25.0 );
            rcsc::AngleDeg l_dir = ( l_limit - wm.self().pos() ).th();
            if ( l_dir.degree() < wm.self().body().degree()
                 && wm.self().body().degree() < goal_dir.degree() )
            {
                already_facing = true;;
            }
        }

        if ( already_facing )
        {
            Line2D body_line( wm.self().pos(), wm.self().body() );
            target_point.y = body_line.getY( target_point.x );
            agent->debugClient().addMessage( "ToBodyDir" );
        }
    }
#else

    rcsc::Vector2D target_point = home_pos;

    if ( wm.ball().pos().y * wm.self().pos().y < 0.0 // opposite side
         //&& wm.ball().pos().x > 45.0 // very chance
         && wm.ball().pos().x > 40.0 // very chance
         && 6.0 < wm.ball().pos().absY() && wm.ball().pos().absY() < 17.0 )
    {
        rcsc::Rect2D goal_area( rcsc::Vector2D( 52.5 - 8.0, -6.0 ),
                                rcsc::Vector2D( 52.5, 6.0 ) );
        if ( ! wm.existTeammateIn( goal_area, 10, true ) )
        {
            agent->debugClient().addMessage( "GoToCenterCross" );

            target_point.x = wm.ball().pos().x + 2.0;
            if ( target_point.x > wm.getOffsideLineX() - 1.0 )
            {
                target_point.x = wm.getOffsideLineX() - 1.0;
            }
            if ( target_point.x > 49.0 ) target_point.x = 49.0;

            target_point.y = -1.0;
            if ( wm.ball().pos().y > 0.0 ) target_point.y *= -1.0;
        }
    }

#endif
    // consider near opponent
    if ( wm.self().pos().dist( target_point ) < 5.0 )
    {
        double opp_dist = 200.0;
        wm.getOpponentNearestTo( target_point, 10, &opp_dist );
        if ( opp_dist < 1.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: ShootAreaMove. Point Blocked. old target=(%.2f, %.2f)"
                                ,__FILE__, __LINE__,
                                target_point.x, target_point.y );
            agent->debugClient().addMessage( "AvoidOpp" );
            if ( target_point.x > wm.getOffsideLineX() - 1.0 )
            {
                target_point.x -= 3.0;
            }
            else
            {
                target_point.x += 3.0;
                if ( target_point.x > wm.getOffsideLineX() - 1.0 )
                {
                    target_point.x = wm.getOffsideLineX() - 1.0;
                }
            }
        }
    }

    // consider goalie
    if ( target_point.x > 45.0 )
    {
        const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
        if ( opp_goalie
             && opp_goalie->distFromSelf() < wm.ball().distFromSelf() )
        {
            rcsc::Line2D cross_line( wm.ball().pos(), target_point );
            if (  cross_line.dist( opp_goalie->pos() ) < 2.0 )
            {
                agent->debugClient().addMessage( "AvoidGK" );
                agent->debugClient().addLine( wm.ball().pos(), target_point );
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s:%d: ShootAreaMove. Goalie is on cross line. old target=(%.2f %.2f)"
                                    ,__FILE__, __LINE__,
                                    target_point.x, target_point.y );
                rcsc::Line2D move_line( wm.self().pos(), target_point );
                target_point.x -= 3.5;
            }
        }
    }

    return target_point;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doShootAreaMove( rcsc::PlayerAgent * agent,
                                  const rcsc::Vector2D& home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doShootAreaMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //------------------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    //------------------------------------------------------

    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doShootAreaMove. intercept cycle. self=%d, mate=%d, opp=%d"
                        ,__FILE__, __LINE__,
                        self_min, mate_min, opp_min );
    if ( self_min < 4
         || self_min < mate_min
         || wm.getInterceptTable()->isSelfFastestPlayer()
         || ( wm.existKickableOpponent() && wm.ball().distFromSelf() < 5.0 ) )
    {
        rcsc::Body_Intercept().execute( agent );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doShootAreaMove. get ball"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "SFW:GetBall(1)" );
        if ( wm.self().pos().x > 30.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return;
    }
    //------------------------------------------------------
    if ( wm.getInterceptTable()->isOurTeamBallPossessor() )
    {
        if ( self_min < 20
             && self_min < mate_min
             && wm.ball().pos().absY() > 19.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: doShootAreaMove. get ball(2)"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "SFW:GetBall(2)" );
            rcsc::Body_Intercept().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return;
        }
    }

    //------------------------------------------------------
    // decide move target point & dash power
    const rcsc::Vector2D target_point = getShootAreaMoveTarget( agent, home_pos );
    const double dash_power = getShootAreaMoveDashPower( agent, target_point );


    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doShootAreaMove. go to cross point (%.2f, %.2f)"
                        ,__FILE__, __LINE__,
                        target_point.x, target_point.y );

    agent->debugClient().addMessage( "ShootMoveP%.0f", dash_power );
    agent->debugClient().setTarget( target_point );


    // move to
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    if ( wm.self().pos().x > target_point.x + dist_thr*0.5
         && std::fabs( wm.self().pos().x - target_point.x ) < 3.0
         && wm.self().body().abs() < 10.0 )
    {
        agent->debugClient().addMessage( "Back" );
        double back_dash_power
            = wm.self().getSafetyDashPower( -dash_power );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doShootAreaMove. Back Move"
                            ,__FILE__, __LINE__ );
        agent->doDash( back_dash_power );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power,
                                 5, // cycle
                                 false, // no back dash
                                 true, // save recovery
                                 20.0 // dir thr
                                 ).execute( agent ) )
    {
        rcsc::Vector2D face_point( rcsc::ServerParam::i().pitchHalfLength(),
                                   wm.self().pos().y );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doShootAreaMove. already target point. turn to (%.2f %.2f)"
                            ,__FILE__, __LINE__,
                            face_point.x, face_point.y );
        rcsc::Body_TurnToPoint( face_point ).execute( agent );
    }

    if ( wm.self().pos().x > 35.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doShootAreaMove. Neck to Goalie or Scan"
                      ,__FILE__, __LINE__ );
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                      "%s:%d: doShootAreaMove. Neck to Ball or Scan"
                      ,__FILE__, __LINE__ );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}
